""" Methods for working with references (e.g., journal articles, books)

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-09-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import JournalArticle, PubMedCentralOpenAccesGraphic
import Bio.Entrez
import dateutil.parser
import ftplib
import glob
import lxml.etree
import os
import requests
import tarfile

__all__ = ['get_reference', 'get_pubmed_central_open_access_graphics']


def get_reference(pubmed_id=None, doi=None, cross_ref_session=requests):
    """ Get data about a reference by its PubMed id and/or DOI

    Args:
        pubmed_id (:obj:`str`, optional): PubMed id
        doi (:obj:`str`, optional): DOI
        session (:obj:`requests.sessions.Session`, optional): requests session

    Returns:
        :obj:`JournalArticle`: data about a reference
    """
    if pubmed_id is None and doi is None:
        raise ValueError('A PubMed id and/or DOI is needed to get information about a reference.')

    pubmed_ref = get_reference_from_pubmed(pubmed_id=pubmed_id, doi=doi)
    if pubmed_ref:
        doi = doi or pubmed_ref.doi

    if doi is not None:
        doi_ref = get_reference_from_crossref(doi, session=cross_ref_session)

    ref = pubmed_ref or doi_ref
    ref.doi = doi
    if doi_ref:
        if len(doi_ref.authors) >= len(ref.authors):
            ref.authors = doi_ref.authors
        ref.journal = doi_ref.journal
        ref.pages = doi_ref.pages

    return ref


def get_reference_from_pubmed(pubmed_id=None, doi=None):
    """ Get data about a reference by its PubMed id

    Args:
        pubmed_id (:obj:`str`, optional): PubMed id
        doi (:obj:`str`, optional): DOI

    Returns:
        :obj:`JournalArticle`: data about a reference
    """
    if pubmed_id:
        handle = Bio.Entrez.esummary(db="pubmed", id=pubmed_id, retmode="xml")
        records = list(Bio.Entrez.parse(handle))
        handle.close()
        if len(records) == 0:
            raise ValueError('`{}` is not a valid PubMed id.'.format(pubmed_id))
        record = records[0]
        doi = str(record.get('DOI', None))
    else:
        handle = Bio.Entrez.esearch(db="pubmed", term=doi, retmode="xml")
        record = Bio.Entrez.read(handle)
        handle.close()

        if len(record['IdList']) == 0:
            return None

        pubmed_id = str(record['IdList'][0])

        handle = Bio.Entrez.esummary(db="pubmed", id=pubmed_id, retmode="xml")
        records = list(Bio.Entrez.parse(handle))
        handle.close()
        if len(records) == 0:
            raise ValueError('`{}` is not a valid PubMed id.'.format(pubmed_id))  # pragma: no cover # shouldn't be reachable
        record = records[0]

    try:
        pub_date = dateutil.parser.parse(record['PubDate'])
        date = '{}-{:02d}-{:02d}'.format(pub_date.year, pub_date.month, pub_date.day)
        year = str(pub_date.year)
    except dateutil.parser._parser.ParserError:
        date = str(record['PubDate'])
        year = date.partition(' ')[0]

    pubmed_central_id = None
    handle = Bio.Entrez.esearch(db="pmc", term=pubmed_id + '[pmid]', retmode="xml")
    pmc_record = Bio.Entrez.read(handle)
    handle.close()
    if pmc_record['IdList']:
        handle = Bio.Entrez.esummary(db="pmc", id=pmc_record['IdList'][0], retmode="xml")
        pmc_records = list(Bio.Entrez.parse(handle))
        handle.close()
        if len(pmc_records) == 0:
            raise ValueError('`{}` is not a valid PubMed Central id.'.format(
                pmc_record['IdList'][0]))  # pragma: no cover # shouldn't be reachable
        pmc_record = pmc_records[0]
        pubmed_central_id = str(pmc_record['ArticleIds']['pmcid'])

    return JournalArticle(
        pubmed_id=pubmed_id,
        pubmed_central_id=pubmed_central_id,
        doi=doi,
        authors=[str(author) for author in record['AuthorList']],
        title=str(record['Title']).strip('.') if record['Title'] else None,
        journal=str(record['FullJournalName']) if record['FullJournalName'] else None,
        volume=str(record['Volume']) if record['Volume'] else None,
        issue=str(record['Issue']) if record['Issue'] else None,
        pages=str(record['Pages']) if record['Pages'] else None,
        year=year,
        date=date,
    )


def get_reference_from_crossref(id, session=requests):
    """ Get data about a reference by its DOI

    Args:
        id (:obj:`str`): DOI
        session (:obj:`requests.sessions.Session`, optional): requests session

    Returns:
        :obj:`JournalArticle`: data about a reference
    """
    response = session.get('https://api.crossref.org/works/' + id)
    response.raise_for_status()
    record = response.json()['message']

    return JournalArticle(
        pubmed_id=None,
        pubmed_central_id=None,
        doi=id,
        authors=[
            author.get('name', (author.get('given', '') + ' ' + author.get('family', '')).strip())
            for author in record['author']
        ],
        title=record['title'][0].strip('.') if record['title'] else None,
        journal=record['container-title'][0],
        volume=record['volume'],
        issue=record.get('journal-issue', {}).get('issue', None),
        pages=record.get('page', record.get('article-number', None)),
        year=str(record['published']['date-parts'][0][0]),
        date='-'.join('{:02d}'.format(date_part) for date_part in record['published']['date-parts'][0][0:3]),
    )


def get_pubmed_central_open_access_graphics(id, dirname, session=requests, max_num_ftp_tries=3):
    """ Get the open access graphics for a publication in PubMed Central

    Args:
        id (:obj:`str`): PubMed Central id
        dirname (:obj:`str`): path to save images
        session (:obj:`requests.session.Session`): requests session

    Returns:
        :obj:`list` of :obj:`PubMedCentralOpenAccesGraphic`: list of graphics
    """
    if not os.path.isdir(dirname):
        os.mkdir(dirname)

    response = session.get('https://www.ncbi.nlm.nih.gov/pmc/utils/oa/oa.fcgi?id=' + id)
    response.raise_for_status()
    root = lxml.etree.fromstring(response.content)
    if root.xpath('/OA/error'):
        return []

    tgz_url = root.xpath("/OA/records/record/link[@format='tgz']")[0].attrib['href']
    tgz_path = tgz_url.partition('ftp://')[2].partition('/')[2]
    tgz_filename = os.path.join(dirname, os.path.basename(tgz_path))

    for iter in range(max_num_ftp_tries):
        if not os.path.isfile(tgz_filename):
            ftp = ftplib.FTP('ftp.ncbi.nlm.nih.gov')
            ftp.login()
            with open(tgz_filename, 'wb') as file:
                ftp.retrbinary('RETR ' + tgz_path, file.write)
            ftp.quit()

        if not os.path.isdir(os.path.join(dirname, id)):
            try:
                with tarfile.open(tgz_filename) as file:
                    file.extractall(path=dirname)
                break
            except tarfile.ReadError:
                os.remove(tgz_filename)

        else:
            break

    nxml_filename = glob.glob(os.path.join(dirname, id, "*.nxml"))[0]
    oa_id = os.path.basename(nxml_filename)[0:-5]

    nxml = lxml.etree.parse(nxml_filename).getroot()
    graphics = []
    for figure in nxml.xpath('*//fig'):
        label = figure.xpath('label')
        if len(label):
            label = label[0]

        caption = figure.xpath('caption')
        if len(caption):
            caption = caption[0]

        graphic = figure.xpath('graphic')
        if len(graphic):
            graphic = graphic[0]

        graphics.append(PubMedCentralOpenAccesGraphic(
            id=oa_id + '/' + figure.attrib['id'],
            label=label.text.strip('.') if label is not None else None,
            caption=''.join([lxml.etree.tostring(child).decode('utf8')
                             for child in caption.getchildren()]) if caption is not None else None,
            filename=os.path.join(dirname, id, graphic.attrib['{{{}}}href'.format(graphic.nsmap['xlink'])] + ".jpg"),
        ))

    return graphics



""" Warnings for reports

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-21
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import BioSimulatorsWarning

__all__ = [
    'RepeatDataSetLabelsWarning',
    'MissingReportMetadataWarning',
    'MissingDataWarning',
    'ExtraDataWarning',
    'CannotExportMultidimensionalTableWarning',
]


class RepeatDataSetLabelsWarning(BioSimulatorsWarning):
    """ Warning that multiple data sets with a report have the same label """
    pass  # pragma: no cover


class MissingReportMetadataWarning(BioSimulatorsWarning):
    """ Warning that an exported file of a report will not or does contain comprehensive metadata about the report
    such as the data type and shape of each data set.
    """
    pass  # pragma: no cover


class MissingDataWarning(BioSimulatorsWarning):
    """ Warning that a file does not contain data for one or more data sets of a report. """
    pass  # pragma: no cover


class ExtraDataWarning(BioSimulatorsWarning):
    """ Warning that a file contains additional data that could not be mapped to a data set of a report. """
    pass  # pragma: no cover


class CannotExportMultidimensionalTableWarning(BioSimulatorsWarning):
    """ Warning that a multidimensional report cannot be exported (e.g., to CSV, Excel, or TSV). """
    pass  # pragma: no cover



""" Methods for reading and writing OMEX Metadata files

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-06-23
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..combine.data_model import CombineArchive, CombineArchiveContentFormatPattern  # noqa: F401
from ..config import get_config, Config  # noqa: F401
from .data_model import (Triple, OmexMetadataOutputFormat, OmexMetadataSchema,
                         BIOSIMULATIONS_ROOT_URI_PATTERN,
                         BIOSIMULATIONS_PREDICATE_TYPES,
                         BIOSIMULATIONS_NAMESPACE_PREFIX_MAP)
from .utils import get_local_combine_archive_content_uri, get_global_combine_archive_content_uri
from .validation import validate_biosimulations_metadata
from lxml import etree
import abc
import collections
import json
import os
import pyomexmeta
import rdflib
import re
import tempfile

__all__ = [
    'read_omex_meta_file',
    'write_omex_meta_file',
    'read_omex_meta_files_for_archive',
    'TriplesOmexMetaReader',
    'TriplesOmexMetaWriter',
    'BiosimulationsOmexMetaReader',
    'BiosimulationsOmexMetaWriter',
]


def read_omex_meta_file(filename_or_filenames, archive=None, working_dir=None, config=None):
    """ Read an OMEX Metadata file

    Args:
        filename_or_filenames (:obj:`str` or :obj:`list` of :obj:`str`): path or paths to OMEX Metadata files
        archive (:obj:`CombineArchive`, optional): parent COMBINE archive
        working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`object`: representation of the OMEX Metadata file in :obj:`schema`
            * nested :obj:`list` of :obj:`str`: nested list of errors with the OMEX Metadata file
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the OMEX Metadata file
    """
    content = None
    errors = []
    warnings = []

    if config is None:
        config = get_config()

    if config.OMEX_METADATA_SCHEMA == OmexMetadataSchema.biosimulations:
        return BiosimulationsOmexMetaReader().run(filename_or_filenames, archive=archive, working_dir=working_dir, config=config)

    elif config.OMEX_METADATA_SCHEMA == OmexMetadataSchema.rdf_triples:
        return TriplesOmexMetaReader().run(filename_or_filenames, archive=archive, working_dir=working_dir, config=config)

    else:
        errors.append(['Schema `{}` is not supported. The following schemas are supported:',
                       [['None']] + sorted([
                           [schema.value] for schema in OmexMetadataSchema.__members__.values()
                       ])])
        return (content, errors, warnings)


def write_omex_meta_file(content, filename, config=None):
    """ Write an OMEX Metadata file

    Args:
        content (:obj:`object`): representation of the OMEX Metadata file in :obj:`schema`
        filename (:obj:`str`): path to save OMEX Metadata file
        config (:obj:`Config`, optional): configuration
    """
    if config is None:
        config = get_config()

    if config.OMEX_METADATA_SCHEMA == OmexMetadataSchema.biosimulations:
        return BiosimulationsOmexMetaWriter().run(content, filename, config=config)

    elif config.OMEX_METADATA_SCHEMA == OmexMetadataSchema.rdf_triples:
        return TriplesOmexMetaWriter().run(content, filename, config=config)

    else:
        msg = 'Schema `{}` is not supported. The following schemas are supported:\n  {}'.format(
            config.OMEX_METADATA_SCHEMA.value if config.OMEX_METADATA_SCHEMA else None,
            '\n  '.join(['None'] + sorted([schema.value for schema in OmexMetadataSchema.__members__.values()])))
        raise NotImplementedError(msg)


def read_omex_meta_files_for_archive(archive, archive_dirname, config=None):
    """ Read all of the OMEX Metadata files in an archive

    Args:
        archive (:obj:`CombineArchive`): COMBINE/OMEX archive
        archive_dirname (:obj:`str`): directory with the content of the archive
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`object`: representation of the OMEX Metadata file in :obj:`schema`
            * nested :obj:`list` of :obj:`str`: nested list of errors with the OMEX Metadata file
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the OMEX Metadata file
    """
    content = []
    errors = []
    warnings = []

    if config is None:
        config = get_config()

    filenames = []
    for item in archive.contents:
        if item.format and re.match(CombineArchiveContentFormatPattern.OMEX_METADATA.value, item.format):
            filenames.append(os.path.join(archive_dirname, item.location))

    if filenames:
        return read_omex_meta_file(filenames, archive=archive, working_dir=archive_dirname, config=config)
    else:
        content = []
        errors = [[(
            'The COMBINE/OMEX does not contain an OMEX Metadata file. '
            'Archives must contain metadata for publication to BioSimulations.'
        )]]
        warnings = []
        return (content, errors, warnings)


class OmexMetaReader(abc.ABC):
    """ Base class for reading OMEX Metadata files """

    @ abc.abstractmethod
    def run(self, filename_or_filenames, archive=None, working_dir=None, config=None):
        """ Read an OMEX Metadata file

        Args:
            filename_or_filenames (:obj:`str` or :obj:`list` of :obj:`str`): path or paths to OMEX Metadata files
            archive (:obj:`CombineArchive`, optional): parent COMBINE archive
            working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`tuple`:

                * :obj:`object`: representation of the OMEX Metadata file
                * nested :obj:`list` of :obj:`str`: nested list of errors with the OMEX Metadata file
                * nested :obj:`list` of :obj:`str`: nested list of warnings with the OMEX Metadata file
        """
        pass  # pragma: no cover

    @classmethod
    def read_rdf(cls, filename, config=None):
        """ Read an RDF file

        Args:
            filename (:obj:`str`): path to the RDF file
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`tuple`:

                * :obj:`pyomexmeta.RDF`: RDF representation of the file
                * nested :obj:`list` of :obj:`str`: nested list of errors with the RDF file
                * nested :obj:`list` of :obj:`str`: nested list of warnings with the RDF file
        """
        if config is None:
            config = get_config()

        rdf = None
        errors = []
        warnings = []

        if not os.path.isfile(filename):
            errors.append(['`{}` is not a file.'.format(filename)])
            return (rdf, errors, warnings)

        pyomexmeta_log_level = pyomexmeta.Logger.get_level()
        pyomexmeta.Logger.clear()
        pyomexmeta.Logger.set_level(pyomexmeta.eLogLevel.warn)

        with open(filename, 'rb') as file:
            line = file.readline()

        temp_filename = None
        if line.startswith(b'<?xml ') and b'?>' in line:
            decl, sep, after_decl = line.partition(b'?>')
            if b' version="1.1"' in decl or b" version='1.1'" in decl:
                decl = decl.replace(b' version="1.1"', b' version="1.0"').replace(b" version='1.1'", b" version='1.0'")
                line = decl + sep + after_decl

                with open(filename, 'rb') as file:
                    lines = file.readlines()

                lines[0] = line

                temp_fid, temp_filename = tempfile.mkstemp()
                os.close(temp_fid)
                with open(temp_filename, 'wb') as file:
                    for line in lines:
                        file.write(line)

                filename = temp_filename

        rdf = pyomexmeta.RDF.from_file(filename, config.OMEX_METADATA_INPUT_FORMAT.value)

        if temp_filename:
            os.remove(temp_filename)

        pyomexmeta.Logger.set_level(pyomexmeta_log_level)

        logger = pyomexmeta.Logger()
        num_messages = len(logger)
        for i_message in range(num_messages):
            message = logger[i_message]
            type = message.get_level()
            message = message.get_message()

            if type in ['warn', 'warning']:
                warnings.append([message])
            else:
                rdf = None
                errors.append([message])

        return (rdf, errors, warnings)

    @classmethod
    def get_rdf_triples(cls, rdf):
        """ Read an RDF file

        Args:
            rdf (:obj:`pyomexmeta.RDF`): RDF representation of the file

        Returns:
            * :obj:`list` of :obj:`Triple`: representation of the OMEX Metadata file as list of triples
        """
        query = "SELECT ?subject ?predicate ?object WHERE { ?subject ?predicate ?object }"
        plain_triples = json.loads(rdf.query_results_as_string(query, 'json'))['results']['bindings']
        triples = []
        for plain_triple in plain_triples:
            subject = cls.make_rdf_node(plain_triple['subject'])
            predicate = cls.make_rdf_node(plain_triple['predicate'])
            object = cls.make_rdf_node(plain_triple['object'])

            triples.append(Triple(
                subject=subject,
                predicate=predicate,
                object=object,
            ))

        return triples

    @classmethod
    def make_rdf_node(cls, node):
        """ Make an RDF node

        Args:
            node (:obj:`dict`): node

        Returns:
            :obj:`rdflib.term.BNode`, :obj:`rdflib.term.Literal`, or :obj:`rdflib.term.URIRef`: node
        """
        if node['type'] == 'bnode':
            return rdflib.term.BNode(node['value'])

        elif node['type'] == 'literal':
            return rdflib.term.Literal(node['value'])

        else:
            return rdflib.term.URIRef(node['value'])


class OmexMetaWriter(abc.ABC):
    """ Base class for writing OMEX Metadata files """

    @ abc.abstractmethod
    def run(self, content, filename, config=None):
        """ Write an OMEX Metadata file

        Args:
            content (:obj:`object`): representation of the OMEX Metadata file
            filename (:obj:`str`): path to save OMEX Metadata file
            config (:obj:`Config`, optional): configuration
        """
        pass  # pragma: no cover


class TriplesOmexMetaReader(OmexMetaReader):
    """ Utility for reading an OMEX Metadata file into a list of triples """

    def run(self, filename_or_filenames, archive=None, working_dir=None, config=None):
        """ Read an OMEX Metadata file into a list of triples

        Args:
            filename_or_filenames (:obj:`str` or :obj:`list` of :obj:`str`): path or paths to OMEX Metadata files
            archive (:obj:`CombineArchive`, optional): parent COMBINE archive
            working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`tuple`:

                * :obj:`list` of :obj:`dict`: representation of the OMEX Metadata file as list of triples
                * nested :obj:`list` of :obj:`str`: nested list of errors with the OMEX Metadata file
                * nested :obj:`list` of :obj:`str`: nested list of warnings with the OMEX Metadata file
        """
        if config is None:
            config = get_config()

        triples = None
        errors = []
        warnings = []

        if isinstance(filename_or_filenames, (tuple, list)):
            filenames = filename_or_filenames
        else:
            filenames = [filename_or_filenames]

        for filename in filenames:
            rdf, temp_errors, temp_warnings = self.read_rdf(filename, config=config)

            if working_dir:
                error_filename = os.path.relpath(filename, working_dir)
            else:
                error_filename = filename

            if temp_errors:
                if isinstance(filename_or_filenames, (tuple, list)):
                    errors.append(['The OMEX Metadata file at location `{}` is invalid.'.format(error_filename), temp_errors])
                else:
                    errors.extend(temp_errors)

            if temp_warnings:
                if isinstance(filename_or_filenames, (tuple, list)):
                    warnings.append(['The OMEX Metadata file at location `{}` may be invalid.'.format(error_filename), temp_warnings])
                else:
                    warnings.extend(temp_warnings)

        if errors:
            return (triples, errors, warnings)

        triples = self.get_rdf_triples(rdf)

        return (triples, errors, warnings)


class TriplesOmexMetaWriter(OmexMetaWriter):
    """ Utility for writing a list of triples to an OMEX Metadata file """

    def run(self, triples, filename, namespaces=None, config=None):
        """ Write a list of triples to an OMEX Metadata file

        Args:
            triples (:obj:`list` of :obj:`Triple`): representation of the OMEX Metadata file as list of triples
            filename (:obj:`str`): path to OMEX Metadata file
            config (:obj:`Config`, optional): configuration
        """
        if config is None:
            config = get_config()

        if config.OMEX_METADATA_OUTPUT_FORMAT == OmexMetadataOutputFormat.turtle:
            graph = rdflib.Graph()
            for prefix, namespace in (namespaces or {}).items():
                graph.namespace_manager.bind(prefix, namespace)
            # graph.namespace_manager.bind('omexLibrary', rdflib.Namespace('http://omex-library.org/'))
            # graph.namespace_manager.bind('identifiers', rdflib.Namespace('http://identifiers.org/'))

            for triple in triples:
                graph.add((triple.subject, triple.predicate, triple.object))

            graph.serialize(filename, format="turtle")

        else:
            def get_uri_namespace_id(uri):
                if '#' in uri:
                    namespace, _, id = uri.rpartition('#')
                    namespace += '#'
                    return (namespace, id)
                elif '/' in uri:
                    namespace, _, id = uri.rpartition('/')
                    namespace += '/'
                    return (namespace, id)
                raise ValueError('URI `{}` does not belong to a namespace'.format(uri))

            namespace_prefix_map = dict(BIOSIMULATIONS_NAMESPACE_PREFIX_MAP)
            namespaces = namespaces or {}
            namespaces['rdf'] = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'

            bnode_ids = collections.OrderedDict()

            for triple in triples:
                namespace, _ = get_uri_namespace_id(str(triple.predicate))
                if namespace in namespace_prefix_map:
                    prefix = namespace_prefix_map[namespace]
                else:
                    prefix = 'ns{}'.format(len(namespaces))
                    namespace_prefix_map[namespace] = prefix
                namespaces[prefix] = namespace

                if isinstance(triple.subject, rdflib.term.BNode) and triple.subject not in bnode_ids:
                    bnode_ids[triple.subject] = '{:07d}'.format(len(bnode_ids))

                if isinstance(triple.object, rdflib.term.BNode) and triple.object not in bnode_ids:
                    bnode_ids[triple.object] = '{:07d}'.format(len(bnode_ids))

            root = etree.Element("{{{}}}RDF".format(namespaces['rdf']), nsmap=namespaces)

            for id, node in bnode_ids.items():
                node = etree.Element("{{{}}}Description".format(namespaces['rdf']), nsmap=namespaces)
                root.append(node)

            for triple in triples:
                subject = etree.Element("{{{}}}Description".format(namespaces['rdf']), nsmap=namespaces)
                if isinstance(triple.subject, rdflib.term.URIRef):
                    subject.attrib["{{{}}}about".format(namespaces['rdf'])] = str(triple.subject)
                else:
                    subject.attrib["{{{}}}nodeID".format(namespaces['rdf'])] = bnode_ids[triple.subject]
                root.append(subject)

                namespace, id = get_uri_namespace_id(str(triple.predicate))
                predicate = etree.Element("{{{}}}{}".format(namespace, id), nsmap=namespaces)

                if isinstance(triple.object, rdflib.term.URIRef):
                    predicate.attrib["{{{}}}resource".format(namespaces['rdf'])] = str(triple.object)
                if isinstance(triple.object, rdflib.term.BNode):
                    predicate.attrib["{{{}}}nodeID".format(namespaces['rdf'])] = bnode_ids[triple.object]
                else:
                    predicate.text = str(triple.object)

                subject.append(predicate)

            etree.ElementTree(root).write(filename,
                                          xml_declaration=True,
                                          encoding="utf-8",
                                          standalone=False,
                                          pretty_print=True)

            if config.OMEX_METADATA_OUTPUT_FORMAT != OmexMetadataOutputFormat.rdfxml:
                rdf = pyomexmeta.RDF.from_file(filename, 'rdfxml')
                if rdf.to_file(filename, config.OMEX_METADATA_OUTPUT_FORMAT.value) != 0:
                    raise RuntimeError('Metadata could not be saved to `{}` in `{}` format.'.format(
                        filename, config.OMEX_METADATA_OUTPUT_FORMAT.value))


class BiosimulationsOmexMetaReader(OmexMetaReader):
    """ Utility for reading the metadata about a COMBINE/OMEX archive in an OMEX Metadata
    file into a dictionary with BioSimulations schema """

    def run(self, filename_or_filenames, archive=None, working_dir=None, config=None):
        """ Read the metadata about a COMBINE/OMEX archive in an OMEX Metadata file into a dictionary
         with BioSimulations schema

        Args:
            filename_or_filenames (:obj:`str` or :obj:`list` of :obj:`str`): path or paths to OMEX Metadata files
            archive (:obj:`CombineArchive`, optional): parent COMBINE archive
            working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`tuple`:

                * :obj:`dict`: representation of the metadata about a COMBINE/OMEX
                    archive in an OMEX Metadata file as a dictionary with BioSimulations schema
                * nested :obj:`list` of :obj:`str`: nested list of errors with the OMEX Metadata file
                * nested :obj:`list` of :obj:`str`: nested list of warnings with the OMEX Metadata file
        """
        if config is None:
            config = get_config()

        el_metadatas = None
        errors = []
        warnings = []

        if isinstance(filename_or_filenames, (tuple, list)):
            filenames = filename_or_filenames
        else:
            filenames = [filename_or_filenames]

        triples = []
        for filename in filenames:
            rdf, temp_errors, temp_warnings = self.read_rdf(filename, config=config)

            if working_dir:
                error_filename = os.path.relpath(filename, working_dir)
            else:
                error_filename = filename

            if temp_errors:
                if isinstance(filename_or_filenames, (tuple, list)):
                    errors.append(['The OMEX Metadata file at location `{}` is invalid.'.format(error_filename), temp_errors])
                else:
                    errors.extend(temp_errors)
            else:
                triples.extend(self.get_rdf_triples(rdf))

            if temp_warnings:
                if isinstance(filename_or_filenames, (tuple, list)):
                    warnings.append(['The OMEX Metadata file at location `{}` may be invalid.'.format(error_filename), temp_warnings])
                else:
                    warnings.extend(temp_warnings)

        if errors:
            return (el_metadatas, errors, warnings)

        combine_archive_uri, temp_errors, temp_warnings = self.get_combine_archive_uri(triples)
        errors.extend(temp_errors)
        warnings.extend(temp_warnings)
        if errors:
            return (el_metadatas, errors, warnings)

        el_metadatas, temp_errors, temp_warnings = self.parse_triples_to_schema(triples, combine_archive_uri)
        errors.extend(temp_errors)
        warnings.extend(temp_warnings)
        if errors:
            return (el_metadatas, errors, warnings)

        temp_errors, temp_warnings = validate_biosimulations_metadata(el_metadatas, archive=archive, working_dir=working_dir)
        errors.extend(temp_errors)
        warnings.extend(temp_warnings)

        return (el_metadatas, errors, warnings)

    @classmethod
    def get_combine_archive_uri(cls, triples):
        """ Get the URI used to the describe the COMBINE/OMEX archive in a list of RDF triples

        Args:
            triples (:obj:`list` of :obj:`dict`): representation of the OMEX Metadata file as list of triples

        Returns:
            :obj:`str`: URI used to the describe the COMBINE/OMEX archive in the list of triples
        """
        archive_uris = set()
        for triple in triples:
            if isinstance(triple.subject, rdflib.term.URIRef):
                archive_uri = re.match(BIOSIMULATIONS_ROOT_URI_PATTERN, str(triple.subject))
                if archive_uri:
                    archive_uris.add(archive_uri.group(1))

        if len(archive_uris) == 0:
            msg = 'File does not contain metadata about an OMEX archive.'
            return (None, [[msg]], [])

        elif len(archive_uris) > 1:
            msg = 'File contains metadata about multiple OMEX archives. File must contain data about 1 archive.'
            return (None, [[msg, [[uri] for uri in archive_uris]]], [])

        else:
            return (list(archive_uris)[0], [], [])

    @classmethod
    def parse_triples_to_schema(cls, triples, combine_archive_uri):
        """ Convert a graph of RDF triples into BioSimulations' metadata schema

        Args:
            triples (:obj:`list` of :obj:`dict`): representation of the OMEX Meta file as list of triples
            combine_archive_uri (:obj:`str`): URI used to the describe the COMBINE/OMEX archive in the list of triples

        Returns:
            :obj:`list` of :obj:`object`: representation of the triples in BioSimulations' metadata schema
        """
        errors = []
        warnings = []

        objects = {}
        for triple in triples:
            for node, is_subject in [(triple.subject, True), (triple.object, False)]:
                object = objects.get(str(node), None)
                if object is None:
                    object = objects[str(node)] = {
                        'type': node.__class__.__name__,
                        'is_subject': is_subject,
                        'is_object': not is_subject,
                    }
                    if isinstance(node, (rdflib.term.BNode, rdflib.term.URIRef)):
                        object['uri'] = str(node)
                    else:
                        object['label'] = str(node)
                object['is_subject'] = object['is_subject'] or is_subject
                object['is_object'] = object['is_object'] or not is_subject

        for triple in triples:
            subject = str(triple.subject)
            predicate = str(triple.predicate)
            object = str(triple.object)

            predicate_type = BIOSIMULATIONS_PREDICATE_TYPES.get(predicate, None)
            if predicate_type is None:
                attr = 'other'
                value = objects[object]
            else:
                attr = predicate_type['attribute']
                value = objects[object]

            if attr not in objects[subject]:
                objects[subject][attr] = []

            objects[subject][attr].append({
                'predicate': predicate,
                'value': value,
            })

        el_metadatas = []
        for uri, raw_metadata in objects.items():
            if (
                raw_metadata['type'] != 'URIRef'
                or raw_metadata['uri'].startswith('local:')
                or not raw_metadata['is_subject']
            ):
                continue

            metadata = {}
            metadata['uri'], metadata['combine_archive_uri'] = get_local_combine_archive_content_uri(uri, combine_archive_uri)
            el_metadatas.append(metadata)
            ignored_statements = []
            for predicate_uri, predicate_type in BIOSIMULATIONS_PREDICATE_TYPES.items():
                metadata[predicate_type['attribute']] = raw_metadata.get(predicate_type['attribute'], [])

                values = []
                for el in metadata[predicate_type['attribute']]:
                    if predicate_type['has_uri'] and predicate_type['has_label']:
                        value = {
                            'uri': None,
                            'label': None,
                        }
                        for sub_el in el['value'].get('other', []):
                            if (
                                sub_el['predicate'] == 'http://dublincore.org/specifications/dublin-core/dcmi-terms/identifier'
                                and 'uri' in sub_el['value']
                            ):
                                value['uri'] = sub_el['value']['uri']
                            elif (
                                sub_el['predicate'] == 'http://www.w3.org/2000/01/rdf-schema#label'
                                and 'label' in sub_el['value']
                            ):
                                value['label'] = sub_el['value']['label']
                        if value['label'] is None:
                            if el['value']['type'] != 'BNode':
                                msg = '({}, {}, {}) does not contain an rdf:label.'.format(
                                    uri, predicate_uri, el['value'].get('uri', None)
                                )
                                ignored_statements.append([msg])
                        else:
                            values.append(value)
                    else:
                        if predicate_type['has_uri']:
                            value = el['value'].get('uri', None)
                            if value is None:
                                msg = '({}, {}) does not contain an URI.'.format(
                                    uri, predicate_uri
                                )
                                ignored_statements.append([msg])
                            else:
                                values.append(value)

                            for sub_el in el['value'].get('other', []):
                                if 'uri' in sub_el['value']:
                                    values.append(sub_el['value']['uri'])

                        else:
                            value = el['value'].get('label', None) or None
                            if value is None:
                                if el['value']['type'] != 'BNode':
                                    msg = '({}, {}, {}) does not contain an rdf:label.'.format(
                                        uri, predicate_uri, el['value'].get('uri', None)
                                    )
                                    ignored_statements.append([msg])
                            else:
                                values.append(value)

                            for sub_el in el['value'].get('other', []):
                                if 'label' in sub_el['value']:
                                    values.append(sub_el['value']['label'])

                metadata[predicate_type['attribute']] = values

                if not predicate_type['multiple_allowed']:
                    if len(metadata[predicate_type['attribute']]) == 0:
                        metadata[predicate_type['attribute']] = None

                    elif len(metadata[predicate_type['attribute']]) == 1:
                        metadata[predicate_type['attribute']] = metadata[predicate_type['attribute']][0]

                    else:
                        metadata[predicate_type['attribute']] = metadata[predicate_type['attribute']][0]
                        msg = 'The COMBINE archive should only contain one instance of predicate `{}`.'.format(
                            predicate_uri
                        )
                        errors.append([msg])

            metadata['other'] = []
            for other_md in raw_metadata.get('other', []):
                value = {
                    'attribute': {
                        'uri': other_md['predicate'],
                        'label': None,
                    },
                    'value': {
                        'uri': None,
                        'label': None
                    },
                }
                for el in other_md['value'].get('description', []):
                    if (
                        el['predicate'] == 'http://dublincore.org/specifications/dublin-core/dcmi-terms/description'
                        and 'label' in el['value']
                    ):
                        value['attribute']['label'] = el['value']['label']
                for el in other_md['value'].get('other', []):
                    if (
                        el['predicate'] == 'http://dublincore.org/specifications/dublin-core/dcmi-terms/identifier'
                        and 'uri' in el['value']
                    ):
                        value['value']['uri'] = el['value']['uri']
                    if (
                        el['predicate'] == 'http://www.w3.org/2000/01/rdf-schema#label'
                        and 'label' in el['value']
                    ):
                        value['value']['label'] = el['value']['label']
                if value['attribute']['label'] is None or value['value']['label'] is None:
                    msg = '({}, {}, {}) does not contain an rdf:label.'.format(
                        uri, other_md['predicate'],
                        other_md['value'].get('label', None)
                        if other_md['value']['type'] == 'Literal'
                        else other_md['value'].get('uri', None)
                    )
                    ignored_statements.append([msg])
                else:
                    metadata['other'].append(value)

            for i_thumbnail, thumbnail in enumerate(metadata['thumbnails']):
                if thumbnail.startswith(combine_archive_uri + '/'):
                    metadata['thumbnails'][i_thumbnail] = './' + thumbnail[len(combine_archive_uri)+1:]
                else:
                    msg = 'Thumbnail URIs must begin with the URI of their parent archive ({}), not `{}`'.format(
                        combine_archive_uri, thumbnail)
                    errors.append([msg])

        if ignored_statements:
            warnings.append(['Some statements were ignored:', ignored_statements])

        if errors:
            return (el_metadatas, errors, warnings)

        return (el_metadatas, errors, warnings)


class BiosimulationsOmexMetaWriter(OmexMetaWriter):
    """ Utility for writing the metadata about a COMBINE/OMEX archive to an OMEX Metadata
    file """

    def run(self, el_metadatas, filename, config=None):
        """ Write an OMEX Metadata file

        Args:
            el_metadatas (:obj:`list` of :obj:`dict`): representation of the metadata about the elements in
                a COMBINE/OMEX archive in an OMEX Metadata file
            filename (:obj:`str`): path to save OMEX Metadata file
            config (:obj:`Config`, optional): configuration
        """
        if config is None:
            config = get_config()

        # convert to triples
        triples = []

        local_id = 0

        namespaces = {
            'dc': rdflib.Namespace('http://dublincore.org/specifications/dublin-core/dcmi-terms/'),
            'dcterms': rdflib.Namespace('http://purl.org/dc/terms/'),
            'foaf': rdflib.Namespace('http://xmlns.com/foaf/0.1/'),
            'rdfs': rdflib.Namespace('http://www.w3.org/2000/01/rdf-schema#'),
        }
        for predicate_type in BIOSIMULATIONS_PREDICATE_TYPES.values():
            namespaces[predicate_type['namespace']['prefix']] = rdflib.Namespace(predicate_type['namespace']['uri'])

        for el_metadata in el_metadatas:
            el_uri = get_global_combine_archive_content_uri(el_metadata['uri'], el_metadata['combine_archive_uri'])

            file_uri_ref = rdflib.term.URIRef(el_uri)
            for predicate_type in BIOSIMULATIONS_PREDICATE_TYPES.values():
                namespace = namespaces[predicate_type['namespace']['prefix']]
                predicate = getattr(namespace, predicate_type['uri'].replace(predicate_type['namespace']['uri'], ''))

                if predicate_type['multiple_allowed']:
                    values = el_metadata.get(predicate_type['attribute'], [])
                else:
                    value = el_metadata.get(predicate_type['attribute'], None)
                    if value is None:
                        values = []
                    else:
                        values = [value]

                for value in values:
                    if predicate_type['has_uri'] and predicate_type['has_label']:
                        local_id += 1
                        object = rdflib.term.URIRef('local:{:05d}'.format(local_id))

                        if value.get('uri', None) is not None:
                            triples.append(Triple(
                                object,
                                namespaces['dc'].identifier,
                                rdflib.term.URIRef(value['uri'])
                            ))

                        if value.get('label', None) is not None:
                            triples.append(Triple(
                                object,
                                namespaces['rdfs'].label,
                                rdflib.term.Literal(value['label'])
                            ))

                        if predicate_type['uri'] in [
                            'http://dublincore.org/specifications/dublin-core/dcmi-terms/creator',
                            'http://dublincore.org/specifications/dublin-core/dcmi-terms/contributor',
                        ]:
                            if value.get('uri', None) is not None:
                                triples.append(Triple(
                                    object,
                                    namespaces['foaf'].accountName,
                                    rdflib.term.URIRef(value['uri']
                                                       .replace('https://identifiers.org/orcid:',
                                                                'https://orcid.org/'))
                                ))

                            if value.get('label', None) is not None:
                                triples.append(Triple(
                                    object,
                                    namespaces['foaf'].name,
                                    rdflib.term.Literal(value['label'])
                                ))

                    elif predicate_type['has_uri']:
                        if predicate_type['uri'] == 'http://www.collex.org/schema#thumbnail':
                            value = get_global_combine_archive_content_uri(value, el_metadata['combine_archive_uri'])

                        object = rdflib.term.URIRef(value)

                    else:
                        object = rdflib.term.Literal(value)

                    triples.append(Triple(
                        subject=file_uri_ref,
                        predicate=predicate,
                        object=object,
                    ))

            for other in el_metadata.get('other', []):
                if '#' in other['attribute']['uri']:
                    namespace, _, predicate = other['attribute']['uri'].partition('#')
                    namespace += '#'
                else:
                    namespace, _, predicate = other['attribute']['uri'].rpartition('/')
                    namespace += '/'

                namespace = rdflib.Namespace(namespace)
                predicate = getattr(namespace, predicate)

                local_id += 1
                object = rdflib.term.URIRef('local:{:05d}'.format(local_id))

                triples.append(Triple(
                    subject=file_uri_ref,
                    predicate=predicate,
                    object=object,
                ))

                if other.get('attribute', {}).get('label', None) is not None:
                    triples.append(Triple(
                        object,
                        namespaces['dc'].description,
                        rdflib.term.Literal(other['attribute']['label'])
                    ))

                if other.get('value', {}).get('uri', None) is not None:
                    triples.append(Triple(
                        object,
                        namespaces['dc'].identifier,
                        rdflib.term.URIRef(other['value']['uri'])
                    ))

                if other.get('value', {}).get('label', None) is not None:
                    triples.append(Triple(
                        object,
                        namespaces['rdfs'].label,
                        rdflib.term.Literal(other['value']['label'])
                    ))

        # save triples to file
        TriplesOmexMetaWriter().run(triples, filename, namespaces=namespaces, config=config)



""" Methods for validating RBA files

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-28
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import os
import rba
import shutil
import tempfile
import zipfile


def validate_model(filename, name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`rba.model.RbaModel`: RBA model
    """
    errors = []
    warnings = []
    model = None

    # check model file exists
    if not isinstance(filename, str):
        errors.append(['`{}` is not a path to an RBA file.'.format(filename)])
        return (errors, warnings, model)

    if not os.path.isfile(filename):
        errors.append(['RBA file `{}` does not exist.'.format(filename)])
        return (errors, warnings, model)

    # check file is a valid zip archive
    temp_dirname = tempfile.mkdtemp()
    try:
        with zipfile.ZipFile(filename, 'r') as zip_file:
            zip_file.extractall(temp_dirname)
    except zipfile.BadZipFile:
        shutil.rmtree(temp_dirname)
        errors.append(['`{}` is not a valid RBA file. RBA files must be zip archives which contain TSV and XML files.'.format(filename)])
        return (errors, warnings, model)

    # check that the file contains a valid RBA model
    try:
        model = rba.RbaModel.from_xml(temp_dirname)
    except Exception as exception:
        shutil.rmtree(temp_dirname)
        errors.append(['`{}` is not a valid RBA file. RBA files must be zip archives which contain TSV and XML files.'.format(
            filename), [[str(exception)]]])
        return (errors, warnings, model)

    shutil.rmtree(temp_dirname)

    return (errors, warnings, model)



""" Utility for converting activity flow diagrams from GINsim Markup Language (GINML) to Vega format

:Author: Andrew Freiburger <afreiburger@uvic.ca>
:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-23
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import copy
import json
import math
import os
import pkg_resources
import xml.etree.ElementTree as et

__all__ = ['ginml_to_vega']


def ginml_to_vega(node_values_data_set, ginml_filename, vega_filename,
                  optimum_nodes_quantity=18,
                  edge_x_value_adjustment=0,
                  edge_y_value_adjustment=45,
                  signal_height=65,
                  signal_padding=10,
                  indent=2):
    """ Convert an activity flow diagram from GINsim Markup Language (GINML) format to Vega format.

    Args:
        node_values_data_set (:obj:`dict`): Vega data set with the predicted values of each node. The dictionary should
            have one of the following schemas

            * Contain a key ``values`` whose value is a list of the predicted node values. Each item of the list
              should be a dictionary with two keys

                * ``label``: value should be the SBML-qual identifier of the corresponding node
                * ``values``: array of the predicted value of the species at each time step

            * Contain a key ``url`` whose value is a URL where the predicted values can be retrieved in JSON
              format. The data should be in the same format outlined above.
            * Contain a key ``sedmlUri`` whose value is an array which contains two elements that indicate

                * The location of the SED-ML file (within its parent COMBINE/OMEX archive) which will produce
                  the predictions for the visualization
                * The id of the SED-ML report which will contain the predictions for the visualization

        ginml_filename (:obj:`str`): path to the visualization in GINML format
        vega_filename (:obj:`str`): path to save the visualization in Vega format
        optimum_nodes_quantity (:obj:`int`)
        edge_x_value_adjustment (:obj:`int`)
        edge_y_value_adjustment (:obj:`int`)
        signal_height (:obj:`int`)
        signal_padding (:obj:`int`)
        indent (:obj:`int` or :obj:`None`, optional): indentation
    """

    # parse the diagram from the GINML file
    with open(ginml_filename, 'r') as file:
        tree = et.parse(file)
    root = tree.getroot()
    graph = root[0]

    nodes = []
    regulations = {}
    for child in graph:
        if child.tag == 'node':
            try:
                comment = child.find('annotation/comment').text
            except AttributeError:
                comment = None

            node_coordinates = child.find('nodevisualsetting/ellipse')
            if node_coordinates is None:
                node_coordinates = child.find('nodevisualsetting/rect')

            y_coordinate = node_coordinates.attrib['y']
            x_coordinate = node_coordinates.attrib['x']
            node_width = node_coordinates.attrib['width']
            node_height = node_coordinates.attrib['height']
            node_fill_color = node_coordinates.attrib['backgroundColor']
            text_color = node_coordinates.attrib['foregroundColor']
            nodes.append({
                'id': child.attrib['id'],
                'name': child.attrib.get('name', child.attrib['id']),
                'x_value': x_coordinate,
                'y_value': y_coordinate,
                'node_width': node_width,
                'node_height': node_height,
                'text_color': text_color,
                'fill_color': node_fill_color,
                'comment': comment
            })

        elif child.tag == 'edge':
            line_points = child.find('edgevisualsetting/polyline')
            edge_color = line_points.attrib['line_color']
            edge_width = line_points.attrib['line_width']
            coordinates = line_points.attrib['points']
            regulations[child.attrib['id']] = {
                'from_node': child.attrib['from'],
                'to_node': child.attrib['to'],
                'sign': child.attrib['sign'],
                'line_points': coordinates,
                'line_color': edge_color,
                'line_width': edge_width
            }

    # the associated nodes for each node are identified and stored
    self_loops = []
    for regulation, value in regulations.items():
        first_node, second_node = regulation.split(':', 1)
        value['associated_nodes'] = [first_node, second_node]
        if first_node == second_node:
            self_loops.append(regulation)

    # determine and parameterize the map coordinates
    min_x = math.inf
    min_y = math.inf
    max_x = -math.inf
    max_y = -math.inf
    for node in nodes:
        min_x = min(min_x, float(node['x_value']))
        max_x = max(max_x, float(node['x_value']))
        min_y = min(min_y, float(node['y_value']))
        max_y = max(max_y, float(node['y_value']))

    ginsim_width = max_x - min_x
    ginsim_height = max_y - min_y

    quantity_nodes = len(nodes)
    # print('Quantity of nodes', quantity_nodes)
    nodes_per_map_scale = (quantity_nodes / optimum_nodes_quantity) ** 0.55
    max_width_height = 700 * nodes_per_map_scale

    if ginsim_width > ginsim_height:
        width = max_width_height
        coordinate_scale = float(width / ginsim_width)
        height = ginsim_height * coordinate_scale
    else:
        height = max_width_height
        coordinate_scale = float(height / ginsim_height)
        width = ginsim_width * coordinate_scale

    width += 160
    height += 50

    # if width > height:
    #     print('Maximum figure dimension', width)
    # else:
    #     print('Maximum figure dimension', height)

    # convert the map nodes to Vega
    net_x_adjustment = - min_x + edge_x_value_adjustment
    net_y_adjustment = - min_y + edge_y_value_adjustment

    node_width_sum = []
    node_height_sum = []
    node_labels = []
    for node in nodes:
        adjusted_x_node = (float(node['x_value']) + net_x_adjustment) * coordinate_scale
        adjusted_y_node = (float(node['y_value']) + net_y_adjustment) * coordinate_scale
        node["node_height"] = float(node["node_height"]) * coordinate_scale
        node["node_width"] = float(node["node_width"]) * coordinate_scale
        node["x_value"] = adjusted_x_node
        node["y_value"] = adjusted_y_node

        node_labels.append({
            'id': node['id'],
            'name': node['name'],
            'text_color': node['text_color'],
            'x': adjusted_x_node + node["node_width"] / 2,
            'y': adjusted_y_node + node["node_height"] / 2
        })

        node_width_sum.append(node["node_width"])
        node_height_sum.append(node["node_height"])

    empirical_width_adjustment = 6 * nodes_per_map_scale ** 0.8
    empirical_height_adjustment = 5 * nodes_per_map_scale ** 0.8
    average_node_width = sum(node_width_sum) / (len(node_width_sum) * 2) + empirical_width_adjustment
    average_node_height = sum(node_height_sum) / (len(node_height_sum) * 2) + empirical_height_adjustment

    # convert regulation paths to Vega
    self_loop_label_y_offset = -33
    edges_label_y_offset = -10 * nodes_per_map_scale
    arc_y_offset = -20 + nodes_per_map_scale ** 2.7
    arc_x_offset = -10
    edge_arc_coordinates = []
    edge_line_coordinates = []
    edge_end_coordinates = []
    regulations_labels = []
    points = []

    for name, value in regulations.items():
        edge_svg_paths = []
        points = value['line_points'].split(' ')
        for point in range(len(points) - 1):
            x1, y1 = points[point].split(',')
            x1 = (float(x1) + net_x_adjustment) * coordinate_scale
            y1 = (float(y1) + net_y_adjustment) * coordinate_scale
            if name in self_loops:
                node_arc_width = 20
                node_arc_radius = 20
                edge_end_x = x1 + arc_x_offset
                edge_end_y = y1 + arc_y_offset
                edge_svg_paths.append({
                    'type': 'arc',
                    'path': "M {} {} A {} {} 0 1 1 {} {}".format(
                        edge_end_x, edge_end_y,
                        node_arc_radius, node_arc_radius,
                        edge_end_x + node_arc_width, edge_end_y,
                    ),
                    'coordinates': [
                        edge_end_x, edge_end_y,
                        node_arc_radius, node_arc_radius,
                        edge_end_x + node_arc_width, edge_end_y,
                    ]
                })
                break

            else:
                if point == 0:
                    x_value = x1
                    y_value = y1
                    cur_x = x1
                    cur_y = y1

                if point + 1 <= len(points) - 1:
                    x2, y2 = points[point + 1].split(',')
                    x2 = (float(x2) + net_x_adjustment) * coordinate_scale
                    y2 = (float(y2) + net_y_adjustment) * coordinate_scale

                    if len(points) == 3:
                        delta_x = x2 - x1
                        delta_y = y2 - y1
                    else:
                        delta_x = x2 - x_value
                        delta_y = y2 - y_value

                    if abs(delta_x) < abs(delta_y):
                        if point + 1 == len(points) - 1:
                            if delta_y < 0:
                                end_offset = (nodes_per_map_scale - 1) * (average_node_height * 0.7)
                            elif delta_y > 0:
                                end_offset = (0.6 - nodes_per_map_scale) * average_node_height * 0.7
                            cur_delta_y = delta_y + end_offset
                        else:
                            cur_delta_y = delta_y
                        edge_svg_paths.append({
                            'type': 'line',
                            'path': 'M {} {} L {} {}'.format(cur_x, cur_y, cur_x, cur_y + cur_delta_y),
                            'coordinates': [cur_x, cur_y, cur_x, cur_y + cur_delta_y],
                        })
                        cur_y += cur_delta_y

                        delta_x = 0
                        edge_end_x = x_value
                        edge_end_y = y_value = y2

                    elif abs(delta_x) > abs(delta_y):
                        if point + 1 == len(points) - 1:
                            if delta_x < 0:
                                end_offset = (nodes_per_map_scale - 1) * (average_node_width / 4)
                            elif delta_x > 0:
                                end_offset = (1 - nodes_per_map_scale) * (average_node_width / 4)
                            cur_delta_x = delta_x + end_offset
                        else:
                            cur_delta_x = delta_x
                        edge_svg_paths.append({
                            'type': 'line',
                            'path': 'M {} {} L {} {}'.format(cur_x, cur_y, cur_x + cur_delta_x, cur_y),
                            'coordinates': [cur_x, cur_y, cur_x + cur_delta_x, cur_y],
                        })
                        cur_x += cur_delta_x

                        delta_y = 0
                        edge_end_y = y_value
                        edge_end_x = x_value = x2

                    else:
                        raise ValueError('The {} edge is inconsistent with a cardinal direction.'.format(name))

                else:
                    raise ValueError('The {} contains an undefined point'.format(name))

                '''if point + 1 == len(points) - 1:
                    if delta_x == 0:
                        edge_end_x = x1
                    elif delta_y == 0:
                        edge_end_y = y1'''

        # establishing the edge ends on the figure
        if value['sign'] == 'positive':
            edge_end_shape = 'triangle'
            edge_end_angle = math.degrees(math.atan2(delta_y, delta_x))
            stroke_width = 1
        elif value['sign'] == 'negative':
            edge_end_shape = 'stroke'
            edge_end_angle = math.degrees(math.atan2(delta_y, delta_x)) + 90
            stroke_width = 5
        elif value['sign'] == 'unknown':
            edge_end_shape = 'circle'
            edge_end_angle = math.degrees(math.atan2(delta_y, delta_x))
            stroke_width = 2

        if delta_x > 0:
            edge_end_x_buffer = -average_node_width
            edge_end_y_buffer = 0
            if edge_end_shape == 'triangle':
                edge_end_angle -= 150
        elif delta_x < 0:
            edge_end_x_buffer = average_node_width
            edge_end_y_buffer = 0
            if edge_end_shape == 'triangle':
                edge_end_angle += 210
        if delta_y > 0:
            edge_end_x_buffer = 0
            edge_end_y_buffer = -average_node_height
            if edge_end_shape == 'triangle':
                edge_end_angle += 90
        elif delta_y < 0:
            edge_end_x_buffer = 0
            edge_end_y_buffer = average_node_height
            if edge_end_shape == 'triangle':
                edge_end_angle += 90

        if name in self_loops:
            labels_y_offset = self_loop_label_y_offset
            edge_ends_y_offset = -5
            edge_ends_x_offset = node_arc_width + 5
            edge_end_angle = -135
        elif name not in self_loops:
            labels_y_offset = edges_label_y_offset
            edge_ends_y_offset = edge_end_y_buffer
            edge_ends_x_offset = edge_end_x_buffer

        regulations_labels.append({'name': name,
                                   'line_color': value['line_color'],
                                   'x': x1,
                                   'y': y1 + labels_y_offset})

        edge_end_coordinates.append({
            'name': name,
            'x': edge_end_x + edge_ends_x_offset,
            'y': edge_end_y + edge_ends_y_offset,
            'edge_end_angle': edge_end_angle,
            'edge_end_shape': edge_end_shape,
            'stroke_width': stroke_width,
            'line_color': value['line_color']
        })

        for edge_svg_path in edge_svg_paths:
            if edge_svg_path['type'] == 'arc':
                edge_arc_coordinates.append({
                    "name": name,
                    'from_node': value['from_node'],
                    'to_node': value['to_node'],
                    'path': edge_svg_path['path'],
                    'x1': edge_svg_path['coordinates'][0],
                    'y1': edge_svg_path['coordinates'][1],
                    'radius': edge_svg_path['coordinates'][2],
                    'x2': edge_svg_path['coordinates'][4],
                    'y2': edge_svg_path['coordinates'][5],
                    'line_color': value['line_color'],
                    'line_width': value['line_width'],
                    'related_nodes': value['associated_nodes']
                })
            else:
                edge_line_coordinates.append({
                    "name": name,
                    'from_node': value['from_node'],
                    'to_node': value['to_node'],
                    'path': edge_svg_path['path'],
                    'x1': edge_svg_path['coordinates'][0],
                    'y1': edge_svg_path['coordinates'][1],
                    'x2': edge_svg_path['coordinates'][2],
                    'y2': edge_svg_path['coordinates'][3],
                    'line_color': value['line_color'],
                    'line_width': value['line_width'],
                    'related_nodes': value['associated_nodes']
                })

    template_filename = pkg_resources.resource_filename(
        'biosimulators_utils',
        os.path.join('viz', 'vega', 'ginml', 'template.json'),
    )
    with open(template_filename, 'r') as file:
        vega = json.load(file)

    vega['width'] = width
    vega['height'] = height + signal_height + signal_padding

    for signal in vega['signals']:
        if signal['name'] == 'MaxX':
            signal['value'] = width

        elif signal['name'] == 'MaxY':
            signal['value'] = height

        elif signal['name'] == 'MinTime':
            signal.pop('value', None)
            signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'outputStartTime']

        elif signal['name'] == 'MaxTime':
            signal.pop('value', None)
            signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'outputEndTime']

        elif signal['name'] == 'Time_step':
            signal.pop('value', None)
            signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'outputStartTime']

            signal['bind']['min'] = {
                'sedmlUri': ['SedDocument:[0]', 'Simulation:[0]', 'outputStartTime'],
            }
            signal['bind']['max'] = {
                'sedmlUri': ['SedDocument:[0]', 'Simulation:[0]', 'outputEndTime'],
            }

        elif signal['name'] == 'Report':
            signal.pop('value', None)
            signal['sedmlUri'] = ['SedDocument:[0]', 'Report:[0]', 'id']

            signal['bind']['options'] = {
                'sedmlUri': ['SedDocument:*', 'Report:*', 'id'],
            }
            signal['bind']['labels'] = {
                'sedmlUri': ['SedDocument:*', 'Report:*', 'name'],
            }

        elif signal['name'] == 'SignalHeight':
            signal['value'] = signal_height

        elif signal['name'] == 'SignalPadding':
            signal['value'] = signal_padding

    for i_entry, entry in enumerate(vega['data']):
        if entry['name'] == 'nodesData':
            entry['values'] = nodes

        elif entry['name'] == 'nodesLabelsData':
            entry['values'] = node_labels

        elif entry['name'] == 'edgesArcData':
            entry['values'] = edge_arc_coordinates

        elif entry['name'] == 'edgesLineData':
            entry['values'] = edge_line_coordinates

        elif entry['name'] == 'edgeEndCoordinatesData':
            entry['values'] = edge_end_coordinates

        elif entry['name'] == 'edgesLabelsData':
            entry['values'] = regulations_labels

        elif entry['name'] == 'nodesValues':
            vega['data'][i_entry] = copy.copy(node_values_data_set)
            vega['data'][i_entry]['name'] = 'nodesValues'

    # save Vega-formatted map
    with open(vega_filename, 'w') as file:
        json.dump(vega, file, indent=indent)



""" Utilities for working with XML documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-13
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import copy
import lxml.etree
import re

__all__ = [
    'get_namespaces_for_xml_doc',
    'get_namespaces_for_xml_element',
    'get_attributes_of_xpaths',
    'validate_xpaths_ref_to_unique_objects',
    'eval_xpath',
    'get_namespaces_with_prefixes',
]


def get_namespaces_for_xml_doc(element_tree):
    """ Get the namespaces used by an XML document

    Args:
        et (:obj:`etree._ElementTree`)

    Returns:
        :obj:`dict`: dictionary that maps the id of each namespace to its URI
    """
    root = element_tree.getroot()
    return get_namespaces_for_xml_element(root)


def get_namespaces_for_xml_element(element_tree):
    """ Get the namespaces used by an XML document

    Args:
        et (:obj:`etree._ElementTree`)

    Returns:
        :obj:`dict`: dictionary that maps the id of each namespace to its URI
    """
    namespaces = copy.copy(element_tree.nsmap)
    if None in namespaces:
        namespaces.pop(None)
        match = re.match(r'^{(.*?)}(.*?)$', element_tree.tag)
        if match:
            namespaces[match.group(2)] = match.group(1)

    nodes_to_check = [(element_tree, element_tree.nsmap)]
    while nodes_to_check:
        node, parent_namespaces = nodes_to_check.pop()

        node_namespaces = copy.copy(node.nsmap)
        for prefix, uri in parent_namespaces.items():
            if node_namespaces.get(prefix, None) == uri:
                node_namespaces.pop(prefix)

        if None in node_namespaces:
            node_namespaces.pop(None)
            match = re.match(r'^{(.*?)}(.*?)$', node.tag)
            if match:
                node_namespaces[match.group(2)] = match.group(1)

        for prefix, uri in node_namespaces.items():
            if prefix in namespaces and namespaces[prefix] != uri:
                msg = 'Prefixes must be used consistently throughout the XML document. Prefix `{}` is not used consistently.'.format(prefix)
                raise ValueError(msg)
            namespaces[prefix] = uri

        for child in node.getchildren():
            nodes_to_check.append((child, node.nsmap))

    return namespaces


def get_attributes_of_xpaths(etree, x_paths, namespaces, attr='id'):
    """ Determine the values of the attributes of the objects that match each XPath

    Args:
        etree (:obj:`lxml.etree._ElementTree`): element tree for XML document
        x_paths (:obj:`list` of `str`): XPaths
        namespaces (:obj:`dict`): dictionary that maps the prefixes of namespaces to their URIs
        attr (:obj:`str` or :obj:`dict`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`list` of :obj:`str`: dictionary that maps each XPath to the
            values of the attribute of the objects in the XML file that match the XPath
    """
    # get namespaces
    if isinstance(attr, dict):
        attr = '{{{}}}{}'.format(attr['namespace']['uri'], attr['name'])

    # determine the values of the attributes of the objects that match each XPath
    x_path_attrs = {}
    for x_path in x_paths:
        try:
            objects = etree.xpath(x_path, namespaces=get_namespaces_with_prefixes(namespaces))

            x_path_attrs[x_path] = [obj.attrib.get(attr, None) for obj in objects]
        except Exception:
            x_path_attrs[x_path] = []

    # return the values of the attributes of the objects that match each XPath
    return x_path_attrs


def validate_xpaths_ref_to_unique_objects(etree, x_paths, namespaces, attr='id'):
    """ Validate that each of a list of XPaths matches a single object in an XML file.

    Args:
        etree (:obj:`lxml.etree._ElementTree`): element tree for XML document
        x_paths (:obj:`list` of `str`): XPaths
        namespaces (:obj:`dict`): dictionary that maps the prefixes of namespaces to their URIs
        attr (:obj:`str` or :obj:`dict`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`str`: dictionary that maps each XPath to the
            value of the attribute of the object in the XML file that matches the XPath

    Raises:
        :obj:`ValueError`: if one or more XPaths matches 0 or multiple objects
    """
    x_path_attr_vals = get_attributes_of_xpaths(etree, x_paths, namespaces, attr=attr)

    errors = []

    no_matches = sorted([x_path for x_path, attr_vals in x_path_attr_vals.items() if len(attr_vals) == 0])
    if no_matches:
        errors.append('XPaths must reference unique objects. The following XPaths do not match any objects:\n  - {}'.format(
            '\n  - '.join(no_matches)))

    multiple_matches = sorted([x_path for x_path, attr_vals in x_path_attr_vals.items() if len(attr_vals) > 1])
    if multiple_matches:
        errors.append('XPaths must reference unique objects. The following XPaths match multiple objects:\n  - {}'.format(
            '\n  - '.join(multiple_matches)))

    if errors:
        raise ValueError('\n\n'.join(errors))

    return {key: vals[0] for key, vals in x_path_attr_vals.items()}


def eval_xpath(element, xpath, namespaces):
    """ Get the object(s) at an XPath

    Args:
        element (:obj:`etree._ElementTree`): element tree
        xpath (:obj:`str`): XPath
        namespaces (:obj:`dict`): dictionary that maps the prefixes of namespaces to their URIs

    Returns:
        :obj:`list` of :obj:`etree._ElementTree`: object(s) at the XPath
    """
    try:
        return element.xpath(xpath, namespaces=get_namespaces_with_prefixes(namespaces))
    except lxml.etree.XPathEvalError as exception:
        if namespaces:
            msg = 'XPath `{}` is invalid with these namespaces:\n  {}\n\n  {}'.format(
                xpath,
                '\n  '.join(' - {}: {}'.format(prefix, uri) for prefix, uri in namespaces.items()),
                exception.args[0].replace('\n', '\n  ')),
        else:
            msg = 'XPath `{}` is invalid without namespaces:\n\n  {}'.format(
                xpath, exception.args[0].replace('\n', '\n  ')),

        exception.args = (msg)
        raise


def get_namespaces_with_prefixes(namespaces):
    """ Get a dictionary of namespaces less namespaces that have no prefix

    Args:
        namespaces (:obj:`dict`): dictionary that maps prefixes of namespaces their URIs

    Returns:
        :obj:`dict`: dictionary that maps prefixes of namespaces their URIs
    """
    if None in namespaces:
        namespaces = dict(namespaces)
        namespaces.pop(None)
    return namespaces



""" Utilities for executing tasks in SED-ML files in COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..archive.io import ArchiveWriter
from ..archive.utils import build_archive_from_paths
from ..config import get_config, Config  # noqa: F401
from ..log.data_model import Status, CombineArchiveLog, StandardOutputErrorCapturerLevel  # noqa: F401
from ..log.utils import init_combine_archive_log, get_summary_combine_archive_log, StandardOutputErrorCapturer
from ..report.data_model import VariableResults, ReportFormat, SedDocumentResults  # noqa: F401
from ..sedml.data_model import (SedDocument, Task, Output, Report, DataSet, Plot2D, Curve,  # noqa: F401
                                Plot3D, Surface, Variable)
from ..utils.core import flatten_nested_list_of_strings
from ..warnings import warn, BioSimulatorsWarning
from .exceptions import CombineArchiveExecutionError, NoSedmlError
from .data_model import CombineArchive
from .io import CombineArchiveReader
from .utils import get_sedml_contents, get_summary_sedml_contents
from .validation import validate
from ..viz.data_model import VizFormat  # noqa: F401
import copy
import datetime
import glob
import os
import tempfile
import shutil
import types  # noqa: F401

__all__ = [
    'exec_sedml_docs_in_archive',
]


def exec_sedml_docs_in_archive(sed_doc_executer, archive_filename, out_dir, apply_xml_model_changes=False,
                               sed_doc_executer_supported_features=(Task, Report, DataSet, Plot2D, Curve, Plot3D, Surface),
                               sed_doc_executer_logged_features=(Task, Report, DataSet, Plot2D, Curve, Plot3D, Surface),
                               log_level=StandardOutputErrorCapturerLevel.c,
                               config=None):
    """ Execute the SED-ML files in a COMBINE/OMEX archive (execute tasks and save outputs)

    Args:
        sed_doc_executer (:obj:`types.FunctionType`): function to execute each SED document in the archive.
            The function must implement the following interface::

                def sed_doc_executer(doc, working_dir, base_out_path, rel_out_path=None,
                             apply_xml_model_changes=False,
                             log=None, log_level=StandardOutputErrorCapturerLevel.c, indent=0, config=None):
                    ''' Execute the tasks specified in a SED document and generate the specified outputs

                    Args:
                        doc (:obj:`SedDocument` of :obj:`str`): SED document or a path to SED-ML file which defines a SED document
                        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

                        out_path (:obj:`str`): path to store the outputs

                            * CSV: directory in which to save outputs to files
                              ``{out_path}/{rel_out_path}/{report.id}.csv``
                            * HDF5: directory in which to save a single HDF5 file (``{out_path}/reports.h5``),
                              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

                        rel_out_path (:obj:`str`, optional): path relative to :obj:`out_path` to store the outputs
                        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file
                        log (:obj:`SedDocumentLog`, optional): execution status of document
                        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
                        indent (:obj:`int`, optional): degree to indent status messages
                        config (:obj:`Config`, optional): BioSimulators common configuration
                    '''

        archive_filename (:obj:`str`): path to COMBINE/OMEX archive
        out_dir (:obj:`str`): path to store the outputs of the archive

            * CSV: directory in which to save outputs to files
              ``{ out_dir }/{ relative-path-to-SED-ML-file-within-archive }/{ report.id }.csv``
            * HDF5: directory in which to save a single HDF5 file (``{ out_dir }/reports.h5``),
              with reports at keys ``{ relative-path-to-SED-ML-file-within-archive }/{ report.id }`` within the HDF5 file

        apply_xml_model_changes (:obj:`bool`): if :obj:`True`, apply any model changes specified in the SED-ML files before
            calling :obj:`task_executer`.
        sed_doc_executer_supported_features (:obj:`list` of :obj:`type`, optional): list of the types of elements that the
            SED document executer supports. Default: tasks, reports, plots, data sets, curves, and surfaces.
        sed_doc_executer_logged_features (:obj:`list` of :obj:`type`, optional): list of the types fo elements which that
            the SED document executer logs. Default: tasks, reports, plots, data sets, curves, and surfaces.
        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
        config (:obj:`Config`): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`SedDocumentResults`: results
            * :obj:`CombineArchiveLog`: log
    """
    if not config:
        config = get_config()

    with StandardOutputErrorCapturer(relay=True, level=log_level, disabled=not config.LOG) as archive_captured:
        verbose = config.VERBOSE

        # initialize status and output
        supported_features = sed_doc_executer_supported_features
        logged_features = sed_doc_executer_logged_features

        if SedDocument not in supported_features:
            supported_features = tuple(list(supported_features) + [SedDocument])

        if SedDocument not in logged_features:
            logged_features = tuple(list(logged_features) + [SedDocument])

        start_time = datetime.datetime.now()

        # create output directory
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)

        # create temporary directory to unpack archive
        archive_tmp_dir = tempfile.mkdtemp()

        try:
            # unpack archive and read metadata
            archive = CombineArchiveReader().run(archive_filename, archive_tmp_dir, config=config)

            # validate archive
            errors, warnings = validate(archive, archive_tmp_dir, config=config)
            if warnings:
                msg = 'The COMBINE/OMEX archive may be invalid.\n  {}'.format(
                    flatten_nested_list_of_strings(warnings).replace('\n', '\n  '))
                warn(msg, BioSimulatorsWarning)

            if errors:
                msg = '`{}` is not a valid COMBINE/OMEX archive.\n  {}'.format(
                    archive_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  '))
                raise ValueError(msg)

            # determine files to execute
            sedml_contents = get_sedml_contents(archive)
            if not sedml_contents:
                msg = "COMBINE/OMEX archive '{}' does not contain any executing SED-ML files".format(archive_filename)
                raise NoSedmlError(msg)

            # print summary of SED documents
            print(get_summary_sedml_contents(archive, archive_tmp_dir))

        except Exception as exception:
            shutil.rmtree(archive_tmp_dir)

            archive = CombineArchive()
            archive_tmp_dir = None

            if config.LOG:
                log = init_combine_archive_log(archive, archive_tmp_dir,
                                               supported_features=supported_features,
                                               logged_features=logged_features)

                log.status = Status.FAILED
                log.out_dir = out_dir
                log.exception = exception
                log.output = archive_captured.get_text()
                log.duration = (datetime.datetime.now() - start_time).total_seconds()
                log.finalize()
                log.export()
            else:
                log = None

            if config.DEBUG:
                raise
            else:
                return (None, log)

        if config.COLLECT_COMBINE_ARCHIVE_RESULTS:
            results = SedDocumentResults()
        else:
            results = None
        if config.LOG:
            log = init_combine_archive_log(archive, archive_tmp_dir,
                                           supported_features=supported_features,
                                           logged_features=logged_features)
            log.status = Status.RUNNING
            log.out_dir = out_dir
            log.export()
        else:
            log = None

        # execute SED-ML files: execute tasks and save output
        exceptions = []
        for i_content, content in enumerate(sedml_contents):
            content_filename = os.path.join(archive_tmp_dir, content.location)
            content_id = os.path.relpath(content_filename, archive_tmp_dir)

            print('Executing SED-ML file {}: {} ...'.format(i_content + 1, content_id))

            if config.LOG:
                doc_log = log.sed_documents[content_id]
                doc_log.status = Status.RUNNING
                doc_log.export()
            else:
                doc_log = None

            with StandardOutputErrorCapturer(relay=verbose, level=log_level, disabled=not config.LOG) as doc_captured:
                doc_start_time = datetime.datetime.now()
                if config.COLLECT_COMBINE_ARCHIVE_RESULTS != config.COLLECT_SED_DOCUMENT_RESULTS:
                    config = copy.copy(config)
                    config.COLLECT_SED_DOCUMENT_RESULTS = config.COLLECT_COMBINE_ARCHIVE_RESULTS

                try:
                    working_dir = os.path.dirname(content_filename)
                    doc_results, _ = sed_doc_executer(
                        content_filename,
                        working_dir,
                        out_dir,
                        os.path.relpath(content_filename, archive_tmp_dir),
                        apply_xml_model_changes=apply_xml_model_changes,
                        log=doc_log,
                        log_level=log_level,
                        indent=1,
                        config=config)
                    if config.COLLECT_COMBINE_ARCHIVE_RESULTS:
                        results[content.location] = doc_results
                    if config.LOG:
                        doc_log.status = Status.SUCCEEDED
                except Exception as exception:
                    if config.DEBUG:
                        raise
                    exceptions.append(exception)
                    if config.LOG:
                        doc_log.status = Status.FAILED
                        doc_log.exception = exception

                # update status
                if config.LOG:
                    doc_log.output = doc_captured.get_text()
                    doc_log.duration = (datetime.datetime.now() - doc_start_time).total_seconds()
                    doc_log.export()

        print('')

        if config.BUNDLE_OUTPUTS:
            print('Bundling outputs ...')

            # bundle CSV files of reports into zip archive
            report_formats = config.REPORT_FORMATS
            archive_paths = [os.path.join(out_dir, '**', '*.' + format.value) for format in report_formats if format != ReportFormat.h5]
            archive = build_archive_from_paths(archive_paths, out_dir)
            if archive.files:
                ArchiveWriter().run(archive, os.path.join(out_dir, config.REPORTS_PATH))

            # bundle PDF files of plots into zip archive
            viz_formats = config.VIZ_FORMATS
            archive_paths = [os.path.join(out_dir, '**', '*.' + format.value) for format in viz_formats]
            archive = build_archive_from_paths(archive_paths, out_dir)
            if archive.files:
                ArchiveWriter().run(archive, os.path.join(out_dir, config.PLOTS_PATH))

        # cleanup temporary files
        print('Cleaning up ...')
        if not config.KEEP_INDIVIDUAL_OUTPUTS:

            report_formats = config.REPORT_FORMATS
            viz_formats = config.VIZ_FORMATS
            path_patterns = (
                [os.path.join(out_dir, '**', '*.' + format.value) for format in report_formats if format != ReportFormat.h5]
                + [os.path.join(out_dir, '**', '*.' + format.value) for format in viz_formats]
            )
            for path_pattern in path_patterns:
                for path in glob.glob(path_pattern, recursive=True):
                    os.remove(path)

            for dir_path, dir_names, file_names in os.walk(out_dir, topdown=False):
                for dir_name in list(dir_names):
                    full_dir_name = os.path.join(dir_path, dir_name)
                    if not os.path.isdir(full_dir_name):
                        dir_names.remove(dir_name)
                    elif not os.listdir(full_dir_name):
                        # not reachable because directory would
                        # have already been removed by the iteration for the directory
                        shutil.rmtree(full_dir_name)  # pragma: no cover
                        dir_names.remove(dir_name)  # pragma: no cover
                if not dir_names and not file_names:
                    shutil.rmtree(dir_path)

        shutil.rmtree(archive_tmp_dir)

        # update status
        if config.LOG:
            log.status = Status.FAILED if exceptions else Status.SUCCEEDED
            log.duration = (datetime.datetime.now() - start_time).total_seconds()
            log.finalize()

            # summarize execution
            print('')
            print('============= SUMMARY =============')
            print(get_summary_combine_archive_log(log))

    # update status
    if config.LOG:
        log.output = archive_captured.get_text()
        log.export()

    # raise exceptions
    if exceptions:
        msg = 'The COMBINE/OMEX did not execute successfully:\n\n  {}'.format(
            '\n\n  '.join(str(exceptions).replace('\n', '\n  ') for exceptions in exceptions))
        exception = CombineArchiveExecutionError(msg)

        if config.LOG:
            log.exception = exception
            log.export()

        if config.DEBUG or not config.LOG:
            raise exception

    # flush log
    if config.LOG:
        log.export()

    # return results and log
    return (results, log)



""" Utilities for working with models referenced by SED documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..combine.data_model import CombineArchive, CombineArchiveContent, CombineArchiveContentFormat
from ..combine.io import CombineArchiveWriter
from .data_model import (  # noqa: F401
    SedDocument, Model, ModelLanguage, ModelLanguagePattern, ModelAttributeChange, Simulation,
    Task, Variable, DataGenerator, Report, DataSet, Plot, Plot2D, Curve, Surface)
from .exceptions import UnsupportedModelLanguageError
from .io import SedmlSimulationWriter
import copy
import os
import re
import shutil
import tempfile
import types  # noqa: F401


__all__ = ['get_parameters_variables_outputs_for_simulation', 'build_combine_archive_for_model']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False,
                                                    model_language_options=None):
    """ Get the possible observables for a simulation of a model

    This method supports the following formats

    * SBML
    * SBML-fbc
    * SBML-qual

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:sbml``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types
        model_language_options (:obj:`dict`, optional): additional options to pass to the methods for individual model formats.
            Dictionary that maps model languages (:obj:`ModelLanguage`) to dictionaries of optional keyword arguments for
            each language

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model

    Raises:
        :obj:`UnsupportedModelLanguageError`: if :obj:`model_language` is not a supported language
    """
    if model_language_options is None:
        model_language_options = {}

    # functions are imported here to only import libraries for required model languages
    if model_language and re.match(ModelLanguagePattern.BNGL.value, model_language):
        from biosimulators_utils.model_lang.bngl.utils import get_parameters_variables_outputs_for_simulation

    elif model_language and re.match(ModelLanguagePattern.CellML.value, model_language):
        from biosimulators_utils.model_lang.cellml.utils import get_parameters_variables_outputs_for_simulation

    elif model_language and re.match(ModelLanguagePattern.LEMS.value, model_language):
        # from biosimulators_utils.model_lang.lems.utils import get_parameters_variables_outputs_for_simulation
        raise UnsupportedModelLanguageError(
            'Models of language `{}` are not supported'.format(model_language))

    elif model_language and re.match(ModelLanguagePattern.RBA.value, model_language):
        from biosimulators_utils.model_lang.rba.utils import get_parameters_variables_outputs_for_simulation

    elif model_language and re.match(ModelLanguagePattern.SBML.value, model_language):
        from biosimulators_utils.model_lang.sbml.utils import get_parameters_variables_outputs_for_simulation

    elif model_language and re.match(ModelLanguagePattern.Smoldyn.value, model_language):
        from biosimulators_utils.model_lang.smoldyn.utils import get_parameters_variables_outputs_for_simulation

    elif model_language and re.match(ModelLanguagePattern.XPP.value, model_language):
        from biosimulators_utils.model_lang.xpp.utils import get_parameters_variables_outputs_for_simulation

    else:
        raise UnsupportedModelLanguageError(
            'Models of language `{}` are not supported'.format(model_language))

    return get_parameters_variables_outputs_for_simulation(
        model_filename, model_language, simulation_type, algorithm_kisao_id=algorithm_kisao_id,
        change_level=change_level, native_ids=native_ids, native_data_types=native_data_types,
        **model_language_options.get(model_language, {}),
    )


def build_combine_archive_for_model(model_filename, model_language, simulation_type, archive_filename, extra_contents=None,
                                    **model_language_options):
    """ Build A COMBINE/OMEX archive with a SED-ML file for a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`ModelLanguage`): model language
        simulation_type (:obj:`type`): subclass of :obj:`Simulation`
        archive_filename (:obj:`str`): path to save COMBINE/OMEX archive
        extra_contents (:obj:`dict` of :obj:`str` -> :obj:`CombineArchiveContent`, optional): dictionary that maps the
            local paths of additional files to include in the COMBINE/OMEX archive to their intended locations within the
            COMBINE/OMEX archive
        **model_language_options: additional options to pass to the methods for individual model formats
    """

    # make a SED-ML document for the model
    params, sims, vars, plots = get_parameters_variables_outputs_for_simulation(
        model_filename, model_language, simulation_type, **model_language_options)

    sedml_doc = SedDocument()
    model = Model(
        id='model',
        source=os.path.basename(model_filename),
        language=model_language.value,
        changes=params,
    )
    sedml_doc.models.append(model)
    for i_sim, sim in enumerate(sims):
        sedml_doc.simulations.append(sim)

        if len(sims) > 1:
            sim_suffix = '_' + str(i_sim)
        else:
            sim_suffix = ''

        task = Task(
            id='task' + sim_suffix,
            model=model,
            simulation=sim,
        )
        sedml_doc.tasks.append(task)

        report = Report(
            id='report' + sim_suffix,
        )
        sedml_doc.outputs.append(report)

        var_data_gen_map = {}

        for var in vars:
            var_copy = copy.copy(var)
            var_copy.id = '{}{}'.format(var_copy.id, sim_suffix)
            var_copy.task = task
            data_gen = DataGenerator(
                id='data_generator{}_{}'.format(sim_suffix, var_copy.id),
                name=var_copy.name,
                variables=[var_copy],
                math=var_copy.id,
            )
            sedml_doc.data_generators.append(data_gen)
            var_data_gen_map[var] = data_gen

            report.data_sets.append(DataSet(
                id='data_set{}_{}'.format(sim_suffix, var_copy.id),
                label=var_copy.id,
                name=var_copy.name,
                data_generator=data_gen,
            ))

        for plot in plots:
            plot_copy = copy.copy(plot)
            plot_copy.id = '{}{}'.format(plot_copy.id, sim_suffix)

            if isinstance(plot, Plot2D):
                plot_copy.curves = []
                for curve in plot.curves:
                    assert len(curve.x_data_generator.variables) == 1
                    assert len(curve.y_data_generator.variables) == 1
                    assert curve.x_data_generator.math == curve.x_data_generator.variables[0].id
                    assert curve.y_data_generator.math == curve.y_data_generator.variables[0].id

                    plot_copy.curves.append(Curve(
                        id='{}{}'.format(curve.id, sim_suffix),
                        name=curve.name,
                        x_data_generator=var_data_gen_map[curve.x_data_generator.variables[0]],
                        y_data_generator=var_data_gen_map[curve.y_data_generator.variables[0]],
                        x_scale=curve.x_scale,
                        y_scale=curve.y_scale,
                    ))

            else:
                plot_copy.surfaces = []
                for surface in plot.surfaces:
                    assert len(curve.x_data_generator.variables) == 1
                    assert len(curve.y_data_generator.variables) == 1
                    assert len(curve.z_data_generator.variables) == 1
                    assert curve.x_data_generator.math == curve.x_data_generator.variables[0].id
                    assert curve.y_data_generator.math == curve.y_data_generator.variables[0].id
                    assert curve.z_data_generator.math == curve.z_data_generator.variables[0].id

                    plot_copy.curves.append(Surface(
                        id='{}{}'.format(curve.id, sim_suffix),
                        name=curve.name,
                        x_data_generator=var_data_gen_map[curve.x_data_generator.variables[0]],
                        y_data_generator=var_data_gen_map[curve.y_data_generator.variables[0]],
                        z_data_generator=var_data_gen_map[curve.z_data_generator.variables[0]],
                        x_scale=curve.x_scale,
                        y_scale=curve.y_scale,
                        z_scale=curve.z_scale,
                    ))

            sedml_doc.outputs.append(plot_copy)

    # make temporary directory for archive
    archive_dirname = tempfile.mkdtemp()
    shutil.copyfile(model_filename, os.path.join(archive_dirname, os.path.basename(model_filename)))

    SedmlSimulationWriter().run(sedml_doc, os.path.join(archive_dirname, 'simulation.sedml'))

    # form a description of the archive
    archive = CombineArchive()
    archive.contents.append(CombineArchiveContent(
        location=os.path.basename(model_filename),
        format=CombineArchiveContentFormat[model_language.name].value,
    ))
    archive.contents.append(CombineArchiveContent(
        location='simulation.sedml',
        format=CombineArchiveContentFormat.SED_ML.value,
    ))
    for local_path, extra_content in (extra_contents or {}).items():
        shutil.copyfile(local_path, os.path.join(archive_dirname, extra_content.location))
        archive.contents.append(extra_content)

    # save archive to file
    CombineArchiveWriter().run(archive, archive_dirname, archive_filename)

    # clean up temporary directory for archive
    shutil.rmtree(archive_dirname)



""" Utilities for simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-30
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config, Config  # noqa: F401
from kisao import AlgorithmSubstitutionPolicy, ALGORITHM_SUBSTITUTION_POLICY_LEVELS

__all__ = [
    'get_algorithm_substitution_policy',
]


def get_algorithm_substitution_policy(config=None):
    """ Get the current algorithm substitution policy based on the value of the ``ALGORITHM_SUBSTITUTION_POLICY``
    environment variable.

    Args:
        config (:obj:`Config`, optional): config

    Returns:
        :obj:`AlgorithmSubstitutionPolicy`: policy

    Raises:
        :obj:`ValueError`: if the value of ``ALGORITHM_SUBSTITUTION_POLICY`` is not the name of a recognized policy
    """
    if not config:
        config = get_config()
    policy_name = config.ALGORITHM_SUBSTITUTION_POLICY
    policy = AlgorithmSubstitutionPolicy.__members__.get(policy_name, None)
    if policy is None:
        msg = (
            '`{}` is not a valid value of `ALGORITHM_SUBSTITUTION_POLICY`. '
            '`ALGORITHM_SUBSTITUTION_POLICY` must have one of the following values:\n  - {}'
        ).format(policy_name or '',
                 '\n  - '.join('`' + policy.value + '`' for policy, _ in
                               sorted(ALGORITHM_SUBSTITUTION_POLICY_LEVELS.items(), key=lambda policy_level: policy_level[1])))
        raise ValueError(msg)
    return policy



""" Data model for XPP simulation methods

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-08
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = ['SIMULATION_METHOD_KISAO_MAP']

SIMULATION_METHOD_KISAO_MAP = {
    'rungekutta': {
        'kisao_id': 'KISAO_0000032',
        'parameters': {}
    },
    'discrete': {
        'kisao_id': 'KISAO_0000029',
        'parameters': {
            'seed': 'KISAO_0000488'
        }
    },
    'euler': {
        'kisao_id': 'KISAO_0000030',
        'parameters': {}
    },
    'backeul': {
        'kisao_id': 'KISAO_0000031',
        'parameters': {}
    },
    'adams': {
        'kisao_id': 'KISAO_0000279',
        'parameters': {}
    },
    'gear': {
        'kisao_id': 'KISAO_0000288',
        'parameters': {
            'dtmin': 'KISAO_0000485',
            'dtmax': 'KISAO_0000467',
            'toler': 'KISAO_0000597'
        }
    },
    '5dp': {
        'kisao_id': 'KISAO_0000087',
        'parameters': {
            'toler': 'KISAO_0000209',
            'atoler': 'KISAO_0000211'
        }
    },
    '83dp': {
        'kisao_id': 'KISAO_0000436',
        'parameters': {
            'toler': 'KISAO_0000209',
            'atoler': 'KISAO_0000211'
        }
    },
    '2rb': {
        'kisao_id': 'KISAO_0000033',
        'parameters': {
            'toler': 'KISAO_0000209',
            'atoler': 'KISAO_0000211',
            'bandlo': 'KISAO_0000480',
            'bandup': 'KISAO_0000479'
        }
    },
    'cvode': {
        'kisao_id': 'KISAO_0000019',
        'parameters': {
            'toler': 'KISAO_0000209',
            'atoler': 'KISAO_0000211',
            'bandlo': 'KISAO_0000480',
            'bandup': 'KISAO_0000479'
        }
    },
    'qualrk': {
        'kisao_id': 'KISAO_0000672',
        'parameters': {
            'toler': 'KISAO_0000597',
            'dtmin': 'KISAO_0000485',
            'dtmax': 'KISAO_0000467',
            'newt_iter': 'KISAO_0000665',
            'newt_tol': 'KISAO_0000565',
            'jac_eps': 'KISAO_0000666'
        }
    },
    'modeuler': {
        'kisao_id': 'KISAO_0000301',
        'parameters': {}
    },
    'ymp': {
        'kisao_id': 'KISAO_0000367',
        'parameters': {}
    },
    'volterra': {
        'kisao_id': 'KISAO_0000664',
        'parameters': {
            'vmaxpts': 'KISAO_0000667'
        }
    },
    'stiff': {
        'kisao_id': 'KISAO_0000668',
        'parameters': {
            'dtmin': 'KISAO_0000485',
            'dtmax': 'KISAO_0000467',
            'toler': 'KISAO_0000597'
        }
    }
}



""" Utilities for creating BioSimulators-compliant command-line interfaces for biosimulation tools.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-04
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
from .._version import __version__ as biosimulators_utils_version
from .data_model import EnvironmentVariable  # noqa: F401
from .environ import ENVIRONMENT_VARIABLES
from kisao._version import __version__ as kisao_version
import cement
import libcombine
try:
    import libsbml
except ModuleNotFoundError:  # pragma: no cover
    libsbml = None
import libsedml
import platform
import sys
import termcolor
import types  # noqa: F401

__all__ = [
    'build_cli',
]


def build_cli(cli_name=None, cli_version=None,
              simulator_name=None, simulator_version=None, simulator_url=None,
              combine_archive_executer=None, environment_variables=None):
    """ Create a BioSimulators-compliant command-line application for a biosimulation tool.

    The command-line application will have two inputs

    * A path to a COMBINE/OMEX archive that describes one or more simulations or one or more models
    * A path to a directory to store the outputs of the execution of the simulations defined in the archive

    The command-line application will also support two additional commands

    * A command for printing help information about the command-line application (`-h`, `--help`)
    * A command for printing version information about the command-line application (`-v`, `--version`)

    Args:
        cli_name (:obj:`str`): name of the command-line program (e.g., `copasi`)
        cli_version (:obj:`str`, optional): version of the command-line application
        simulator_name (:obj:`str`): name of the simulator (e.g., `COPASI`)
        simulator_version (:obj:`str`, optional): version of the simulator
        simulator_url (:obj:`str`, optional): URL for information about the simulator
        combine_archive_executer (:obj:`types.FunctionType`): a function which has two positional arguments
            * The path to the COMBINE/OMEX archive
            * The path to the directory to save the outputs of the simulations defined in the archive
        environment_variables (:obj:`list` of :obj:`EnvironmentVariable`, optional): description of the environment
            variables recognized by the simulator

    Returns:
        :obj:`cement.App`: command-line application

    Raises:
        :obj:`SystemExit`: if the execution of the COMBINE/OMEX archive fails
    """
    python_version = '.'.join(str(i) for i in sys.version_info[0:3])

    if not cli_name:
        raise ValueError('CLI name must be defined')
    if not simulator_name:
        raise ValueError('Simulator name must be defined')
    if not combine_archive_executer:
        raise ValueError('COMBINE/OMEX archive execution method must be provided')

    description_value = "BioSimulators-compliant command-line interface to the {} simulation program{}.".format(
        simulator_name,
        ' <{}>'.format(simulator_url) if simulator_url else '')

    if environment_variables is None:
        environment_variables = ENVIRONMENT_VARIABLES.values()

    if environment_variables:
        description_value += ('\n\nIn addition to the command-line arguments outlined below, '
                              '{} also supports the following environment variables:'
                              ).format(simulator_name)

        for env_var in sorted(environment_variables, key=lambda var: var.name):
            description_value += "\n\n  '{}': {}".format(env_var.name, env_var.description)
            if env_var.options:
                option_values = ["'" + val + "'" for val in sorted(env_var.options)]
                description_value += "\n    Options: {}".format(', '.join(option_values))
            if env_var.default:
                description_value += "\n    Default value: '{}'".format(env_var.default)
            if env_var.more_info_url:
                description_value += "\n    More information: {}".format(env_var.more_info_url)

    versions = []
    if simulator_version:
        versions.append(simulator_name + ': ' + simulator_version)
    if cli_version:
        versions.append('CLI: ' + cli_version)
    versions.append('BioSimulators utils: ' + biosimulators_utils_version)
    versions.append('libSED-ML: ' + libsedml.__version__)
    versions.append('KiSAO: ' + kisao_version)
    if libsbml:
        versions.append('libSBML: ' + libsbml.__version__)
    versions.append('libCOMBINE: ' + libcombine.__version__)
    versions.append('Python: ' + python_version)
    versions.append('OS: {} {}'.format(
        platform.system(),
        platform.release(),
    ))
    versions.append('Machine: ' + platform.machine())
    version = '\n'.join(versions)

    class BaseController(cement.Controller):
        """ Base controller for command line application """

        class Meta:
            label = 'base'
            description = description_value
            help = cli_name
            arguments = [
                (
                    ['-i', '--archive'],
                    dict(
                        type=str,
                        required=True,
                        help='Path to COMBINE/OMEX file which contains one or more SED-ML-encoded simulation experiments',
                    ),
                ),
                (
                    ['-o', '--out-dir'],
                    dict(
                        type=str,
                        default='.',
                        help='Directory to save outputs',
                    ),
                ),
                (
                    ['-v', '--version'],
                    dict(
                        action='version',
                        version=version,
                    ),
                ),
            ]

        @cement.ex(hide=True)
        def _default(self):
            args = self.app.pargs
            config = get_config()
            config.LOG = True
            try:
                _, log = combine_archive_executer(args.archive, args.out_dir, config=config)
            except Exception as exception:
                if config.DEBUG:
                    raise
                raise SystemExit(termcolor.colored(str(exception), 'red')) from exception

            if log and log.exception:
                raise SystemExit(termcolor.colored(str(log.exception), 'red')) from log.exception

    class App(cement.App):
        """ Command line application """
        class Meta:
            label = cli_name
            base_controller = 'base'
            handlers = [
                BaseController,
            ]

    return App



""" Utilities for working with XPP models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-08
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (  # noqa: F401
    SedDocument, ModelAttributeChange, Variable, Symbol,
    Simulation, OneStepSimulation, UniformTimeCourseSimulation,
    Algorithm, AlgorithmParameterChange,
    DataGenerator, Plot2D, Plot3D, Curve, Surface, AxisScale,
    Task,
    )
from ...utils.core import flatten_nested_list_of_strings
from .data_model import SIMULATION_METHOD_KISAO_MAP
from .validation import validate_model
import types  # noqa: F401

__all__ = ['get_parameters_variables_outputs_for_simulation']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    set_filename=None, parameter_filename=None, initial_conditions_filename=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file or directory of XPP ODE, set, parameters, and initial conditions
            files
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:xpp``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        set_filename (:obj:`str`, optional): path to XPP set file
        parameter_filename (:obj:`str`, optional): path to XPP parameters file
        initial_conditions_filename (:obj:`str`, optional): path to XPP initial conditions file
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulations of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    # check model file exists and is valid
    errors, _, model = validate_model(model_filename,
                                      set_filename=set_filename,
                                      parameter_filename=parameter_filename,
                                      initial_conditions_filename=initial_conditions_filename)
    if errors:
        raise ValueError('Model file `{}` is not a valid XPP file or directory of XPP files.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))

    if simulation_type not in [OneStepSimulation, UniformTimeCourseSimulation]:
        raise NotImplementedError('`simulation_type` must be `OneStepSimulation` or `UniformTimeCourseSimulation`')

    # parameters
    params = []

    for key, val in model['parameters'].items():
        params.append(ModelAttributeChange(
            id=key if native_ids else 'parameter_{}'.format(key),
            name=None if native_ids else 'Value of parameter "{}"'.format(key),
            target='{}'.format(key),
            new_value=val if native_data_types else str(val),
        ))

    for key, val in model['initial_conditions'].items():
        params.append(ModelAttributeChange(
            id=key if native_ids else 'initial_condition_{}'.format(key),
            name=None if native_ids else 'Initial condition of "{}"'.format(key),
            target='{}'.format(key),
            new_value=val if native_data_types else str(val),
        ))

    # simulation
    sim_method_id = model['simulation_method'].get('meth', 'rungekutta').lower()
    sim_method_props = SIMULATION_METHOD_KISAO_MAP[sim_method_id]
    sim_method_kisao_id = sim_method_props['kisao_id']

    sim = UniformTimeCourseSimulation(
        id='simulation',
        algorithm=Algorithm(
            kisao_id=sim_method_kisao_id,
        )
    )

    t_0 = float(model['simulation_method'].get('t0', 0.))
    duration = float(model['simulation_method'].get('total', 20.))
    d_t = float(model['simulation_method'].get('dt', 0.05))
    n_jmp = float(model['simulation_method'].get('njmp', 1))

    sim.initial_time = t_0
    sim.output_start_time = t_0
    sim.output_end_time = sim.initial_time + duration
    sim.number_of_steps = duration / (d_t * n_jmp)
    sim.number_of_steps = round(sim.number_of_steps)

    for key, val in model['simulation_method'].items():
        param_kisao_id = sim_method_props['parameters'].get(key, None)
        if param_kisao_id:
            sim.algorithm.changes.append(AlgorithmParameterChange(kisao_id=param_kisao_id,
                                                                  new_value=float(val) if native_data_types else val))

    # observables
    vars = []

    vars.append(Variable(
        id=None if native_ids else 'time',
        name=None if native_ids else 'Time',
        symbol=Symbol.time,
    ))

    for key in model['initial_conditions'].keys():
        var = Variable(
            id=key if native_ids else 'dynamics_{}'.format(key),
            name=None if native_ids else 'Dynamics of "{}"'.format(key),
            target=key,
        )
        vars.append(var)

    for key in model['auxiliary_variables'].keys():
        var = Variable(
            id=key if native_ids else 'dynamics_{}'.format(key),
            name=None if native_ids else 'Dynamics of "{}"'.format(key),
            target=key,
        )
        vars.append(var)

    # plots
    if 'elements' in model['plot']:
        plot_type = Plot2D
        for i_element in sorted(model['plot']['elements'].keys()):
            element = model['plot']['elements'][i_element]
            if 'z' in element:
                plot_type = Plot3D
                break

        for i_element in sorted(model['plot']['elements'].keys()):
            element = model['plot']['elements'][i_element]
            if 'x' not in element:
                raise ValueError('Plot element {} must have x data.'.format(i_element + 1))

            if 'y' not in element:
                raise ValueError('Plot element {} must have y data.'.format(i_element + 1))

            if plot_type == Plot3D:
                if 'z' not in element:
                    raise ValueError('3D plot element {} must have z data.'.format(i_element + 1))

        data_generators = {}
        for var in vars:
            data_generators[var.target] = DataGenerator(
                id='data_generator_{}'.format(var.target),
                name=var.target,
                variables=[var],
                math=var.id,
            )

        plot = plot_type(id='plot')

        for i_element in sorted(model['plot']['elements'].keys()):
            element = model['plot']['elements'][i_element]

            if plot_type == Plot2D:
                plot.curves.append(Curve(
                    id='curve_{}'.format(i_element),
                    name='{} vs {}'.format(element['y'], element['x']),
                    x_data_generator=data_generators[element['x']],
                    y_data_generator=data_generators[element['y']],
                    x_scale=AxisScale.linear,
                    y_scale=AxisScale.linear,
                ))

            else:
                plot.surfaces.append(Surface(
                    id='surface_{}'.format(i_element),
                    name='{} vs {} vs {}'.format(element['z'], element['y'], element['x']),
                    x_data_generator=data_generators[element['x']],
                    y_data_generator=data_generators[element['y']],
                    z_data_generator=data_generators[element['z']],
                    x_scale=AxisScale.linear,
                    y_scale=AxisScale.linear,
                    z_scale=AxisScale.linear,
                ))

        outputs = [plot]
    else:
        outputs = []

    return (params, [sim], vars, outputs)



""" Utilities for working with RBA models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-28
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (  # noqa: F401
    SedDocument, ModelAttributeChange, Variable,
    Simulation, SteadyStateSimulation,
    Algorithm,
    Task,
    )
from ...utils.core import flatten_nested_list_of_strings
from .validation import validate_model
import rba
import types  # noqa: F401

__all__ = ['get_parameters_variables_outputs_for_simulation']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:rba``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000669``
            for RBA)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulations of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    # check model file exists and is valid
    errors, _, model = validate_model(model_filename)
    if errors:
        raise ValueError('Model file `{}` is not a valid RBA file.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))

    if simulation_type not in [SteadyStateSimulation]:
        raise NotImplementedError('`simulation_type` must be `SteadyStateSimulation`')

    # parameters
    params = []
    for function in model.parameters.functions:
        for parameter in function.parameters:
            params.append(ModelAttributeChange(
                id='{}.{}'.format(function.id, parameter.id) if native_ids else 'parameter_{}_{}'.format(function.id, parameter.id),
                name=None if native_ids else 'Value of parameter "{}" of function "{}"'.format(parameter.id, function.id),
                target='parameters.functions.{}.parameters.{}'.format(function.id, parameter.id),
                new_value=parameter.value if native_data_types else str(parameter.value),
            ))

    # simulation
    sim = SteadyStateSimulation(
        id='simulation',
        algorithm=Algorithm(
            kisao_id='KISAO_0000669',
        )
    )

    # observables
    vars = []

    vars.append(Variable(
        id=None if native_ids else 'objective',
        name=None if native_ids else 'Value of objective',
        target='objective',
    ))

    constraint_matrix = rba.ConstraintMatrix(model)

    if set(constraint_matrix.col_names).intersection(set(constraint_matrix.row_names)):
        variable_prefix = 'primal_'
        constraint_prefix = 'dual_'
    else:
        variable_prefix = ''
        constraint_prefix = ''

    for name in constraint_matrix.col_names:
        vars.append(Variable(
            id=name if native_ids else variable_prefix + name,
            name=None if native_ids else 'Primal of variable "{}"'.format(name),
            target='variables.{}'.format(name),
        ))

    for name in constraint_matrix.row_names:
        vars.append(Variable(
            id=name if native_ids else constraint_prefix + name,
            name=None if native_ids else 'Dual of constraint "{}"'.format(name),
            target='constraints.{}'.format(name),
        ))

    return (params, [sim], vars, [])



""" Standard exceptions issued by biosimulation tools

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..exceptions import BioSimulatorsException

__all__ = [
    'AlgorithmDoesNotSupportModelFeatureException',
]


class AlgorithmDoesNotSupportModelFeatureException(BioSimulatorsException):
    """ Exception that an algorithm does not support a feature of a model """
    pass  # pragma: no cover



""" Utilities for validating CellML models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-05-10
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import libcellml
import lxml.etree
import os
import pkg_resources


def validate_model(filename, name=None, resolve_imports=True):
    """ Check that a file is a valid CellML model

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages
        resolve_imports (:obj:`bool`, optional): whether to resolve imports

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`tuple`:

                * :obj:`libcellml.model.Model`: model
                * :obj:`lxml.etree._ElementTree`: model
    """
    errors = []
    warnings = []
    model = None

    if not os.path.isfile(filename):
        errors.append(['`{}` is not a file.'.format(filename)])
        return (errors, warnings, (model, None))

    try:
        doc = lxml.etree.parse(filename)
    except lxml.etree.XMLSyntaxError as exception:
        errors.append(['`{}` is not a valid XML file.'.format(filename), [[str(exception)]]])
        return (errors, warnings, (model, None))

    root = doc.getroot()
    default_ns = root.nsmap.get(None, '')
    if default_ns.startswith('http://www.cellml.org/cellml/1.0'):
        v_errors, v_warnings, model = validate_model_version_1_0(filename, doc, resolve_imports=resolve_imports)

    elif default_ns.startswith('http://www.cellml.org/cellml/1.1'):
        v_errors, v_warnings, model = validate_model_version_1_1(filename, doc, resolve_imports=resolve_imports)

    elif default_ns.startswith('http://www.cellml.org/cellml/2'):
        v_errors, v_warnings, model = validate_model_version_2(filename, doc, resolve_imports=resolve_imports)

    else:
        v_errors = [[(
            '`{}` could not be validated. The default namespace must be the namespace of a version of a CellML, not `{}`.'
        ).format(filename, default_ns)]]
        v_warnings = []

    errors.extend(v_errors)
    warnings.extend(v_warnings)

    return (errors, warnings, (model, root))


def validate_model_version_1_0(filename, doc, resolve_imports=True):
    """ Check that a file is a valid CellML 2.0 model

    Args:
        filename (:obj:`str`): path to model
        doc (:obj:`lxml.etree._ElementTree`): XML document for file
        resolve_imports (:obj:`bool`, optional): whether to resolve imports

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`None`
    """

    # ``cellml1.0.rnc`` and ``mathml2.rnc`` were obtained from https://www.cellml.org/tools/validation
    # ``cellml1.0.rng`` and ``mathml2.rng`` were created from ``cellml1.0.rnc`` and and ``mathml2.rnc`` with:
    #
    #     ```
    #     java -jar /home/jonrkarr/.local/lib/python3.9/site-packages/jingtrang/trang.jar \
    #         biosimulators_utils/model_lang/cellml/cellml1.0.rnc \
    #         biosimulators_utils/model_lang/cellml/cellml1.0.rng
    #     java -jar /home/jonrkarr/.local/lib/python3.9/site-packages/jingtrang/trang.jar \
    #         biosimulators_utils/model_lang/cellml/mathml2.rnc \
    #         biosimulators_utils/model_lang/cellml/mathml2.rng
    #     ```

    errors = []
    warnings = []

    schema_filename = pkg_resources.resource_filename('biosimulators_utils',
                                                      os.path.join("model_lang", "cellml", "cellml1.0.rng"))
    schema_doc = lxml.etree.parse(schema_filename)
    schema = lxml.etree.RelaxNG(schema_doc)

    if not schema.validate(doc):
        for error in schema.error_log.filter_levels([lxml.etree.ErrorLevels.ERROR, lxml.etree.ErrorLevels.FATAL]):
            errors.append([
                '{}.{}: {}'.format(error.line, error.column, error.message)
            ])

        for warning in schema.error_log.filter_levels(lxml.etree.ErrorLevels.WARNING):
            warnings.append([
                '{}.{}: {}'.format(warning.line, warning.column, warning.message)
            ])

    return (errors, warnings, None)


def validate_model_version_1_1(filename, doc, resolve_imports=True):
    """ Check that a file is a valid CellML 2.0 model

    Args:
        filename (:obj:`str`): path to model
        doc (:obj:`lxml.etree._ElementTree`): XML document for file
        resolve_imports (:obj:`bool`, optional): whether to resolve imports

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`None`
    """
    errors = []
    warnings = []

    warnings.append(['Validation is not available for CellML 1.1 files.'])

    return (errors, warnings, None)


def validate_model_version_2(filename, doc, resolve_imports=True):
    """ Check that a file is a valid CellML 2.0 model

    Args:
        filename (:obj:`str`): path to model
        doc (:obj:`lxml.etree._ElementTree`): XML document for file
        resolve_imports (:obj:`bool`, optional): whether to resolve imports

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`libcellml.model.Model`: model
    """
    errors = []
    warnings = []

    # read model
    parser = libcellml.Parser()
    with open(filename, 'r') as file:
        model = parser.parseModel(file.read())

    for i_error in range(parser.errorCount()):
        error = parser.error(i_error)
        errors.append([error.description()])

    for i_warning in range(parser.warningCount()):
        warning = parser.warning(i_warning)
        warnings.append([warning.description()])

    if errors:
        return (errors, warnings, model)

    # import imported models
    if resolve_imports:
        importer = libcellml.Importer()
        if not importer.resolveImports(model, os.path.dirname(filename) + os.path.sep):
            for i_error in range(importer.errorCount()):
                error = importer.error(i_error)
                errors.append([error.description()])

            for i_warning in range(importer.warningCount()):
                warning = importer.warning(i_warning)
                warnings.append([warning.description()])

        if errors:
            return (errors, warnings, model)

    # validate model
    validator = libcellml.Validator()
    validator.validateModel(model)

    for i_error in range(validator.errorCount()):
        error = validator.error(i_error)
        errors.append([error.description()])

    for i_warning in range(validator.warningCount()):
        warning = validator.warning(i_warning)
        warnings.append([warning.description()])

    return (errors, warnings, model)



""" Utilities for working with COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..sedml.data_model import Report, Plot, Plot2D, Plot3D
from ..sedml.io import SedmlSimulationReader
from .data_model import CombineArchive, CombineArchiveContent, CombineArchiveContentFormatPattern  # noqa: F401
import os
import re

__all__ = ['get_sedml_contents', 'get_summary_sedml_contents']


def get_sedml_contents(archive,
                       include_all_sed_docs_when_no_sed_doc_is_master=True,
                       always_include_all_sed_docs=False):
    """ Get the SED-ML files in an archive

    Args:
        archive (:obj:`CombineArchive`): COMBINE/OMEX archive
        include_all_sed_docs_when_no_sed_doc_is_master (:obj:`bool`, optional): if :obj:`true`
            and no SED document has ``master="true"``, return all SED documents.
        always_include_all_sed_docs (:obj:`bool`, optional): if :obj:`true`,
            return all SED documents, regardless of whether they have ``master="true"`` or not.

    Returns:
        :obj:`list` of :obj:`CombineArchiveContent`: SED-ML files in a COMBINE/OMEX archive
    """

    sedml_contents = []
    master_sedml_contents = []
    for content in archive.contents:
        if (
            isinstance(content, CombineArchiveContent)
            and content.format
            and re.match(CombineArchiveContentFormatPattern.SED_ML.value, content.format)
        ):
            sedml_contents.append(content)
            if content.master:
                master_sedml_contents.append(content)

    if always_include_all_sed_docs:
        return sedml_contents
    else:
        if master_sedml_contents:
            return master_sedml_contents
        else:
            if include_all_sed_docs_when_no_sed_doc_is_master:
                return sedml_contents
            else:
                return []


def get_summary_sedml_contents(archive, archive_dir,
                               include_all_sed_docs_when_no_sed_doc_is_master=True,
                               always_include_all_sed_docs=False):
    """ Get a summary of the SED-ML content in a COMBINE/OMEX archive

    Args:
        archive (:obj:`CombineArchive`): COMBINE/OMEX archive
        archive_dir (:obj:`str`): path where the content of the archive is located
        include_all_sed_docs_when_no_sed_doc_is_master (:obj:`bool`, optional): if :obj:`true`
            and no SED document has ``master="true"``, return all SED documents.
        always_include_all_sed_docs (:obj:`bool`, optional): if :obj:`true`,
            return all SED documents, regardless of whether they have ``master="true"`` or not.

    Returns:
        :obj:`str`: summary of the SED-ML content in a COMBINE/OMEX archive
    """
    contents = get_sedml_contents(archive,
                                  include_all_sed_docs_when_no_sed_doc_is_master=include_all_sed_docs_when_no_sed_doc_is_master,
                                  always_include_all_sed_docs=always_include_all_sed_docs)

    n_docs = 0
    n_models = 0
    n_simulations = 0
    n_tasks = 0
    n_reports = 0
    n_plots = 0
    for i_content, content in enumerate(contents):
        n_docs += 1

        content_filename = os.path.join(archive_dir, content.location)
        doc = SedmlSimulationReader().run(content_filename, validate_models_with_languages=False)

        n_models += len(doc.models)
        n_simulations += len(doc.simulations)
        n_tasks += len(doc.tasks)
        for output in doc.outputs:
            if isinstance(output, Report):
                n_reports += 1
            else:
                n_plots += 1
    summary = 'Archive contains {} SED-ML documents with {} models, {} simulations, {} tasks, {} reports, and {} plots:\n'.format(
        n_docs, n_models, n_simulations, n_tasks, n_reports, n_plots)

    for i_content, content in enumerate(sorted(contents, key=lambda content: content.location)):
        content_filename = os.path.join(archive_dir, content.location)
        doc = SedmlSimulationReader().run(content_filename, validate_models_with_languages=False)
        content_id = os.path.relpath(content_filename, archive_dir)
        summary += '  {}:\n'.format(content_id)

        sorted_tasks = sorted(doc.tasks, key=lambda task: task.id)
        sorted_reports = sorted((output for output in doc.outputs if isinstance(output, Report)), key=lambda output: output.id)
        sorted_plots = sorted((output for output in doc.outputs if isinstance(output, Plot)), key=lambda output: output.id)

        if sorted_tasks:
            summary += '    Tasks ({}):\n'.format(len(sorted_tasks))
            for task in sorted_tasks:
                summary += '      {}\n'.format(task.id)

        if sorted_reports:
            summary += '    Reports ({}):\n'.format(len(sorted_reports))
            for output in sorted_reports:
                summary += '      {}: {} data sets\n'.format(output.id, len(output.data_sets))

        if sorted_plots:
            summary += '    Plots ({}):\n'.format(len(sorted_plots))
            for output in sorted_plots:
                if isinstance(output, Plot2D):
                    summary += '      {}: {} curves\n'.format(output.id, len(output.curves))
                elif isinstance(output, Plot3D):
                    summary += '      {}: {} surfaces\n'.format(output.id, len(output.surfaces))

    return summary



""" Data model for simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum

__all__ = [
    'SoftwareInterface',
    'EnvironmentVariable',
]


class SoftwareInterface(str, enum.Enum):
    """ Software interface """
    library = 'library'
    command_line = 'command-line application'
    desktop_application = 'desktop application'
    mobile_application = 'mobile application'
    web_service = 'web service'
    web_application = 'web application'
    biosimulators_docker_image = 'BioSimulators Docker image'


class EnvironmentVariable(object):
    """ Environment variable supported by a simulator

    Attributes:
        name(: obj: `str`): name
        description(: obj: `str`): description
        options(: obj: `type`): enumeration of options
        default(: obj: `str`): default
        more_info_url(: obj: `str`): URL with more information about the variable
    """

    def __init__(self, name=None, description=None, options=None, default=None, more_info_url=None):
        """
        Args:
            name(: obj: `str`, optional): name
            description(: obj: `str`, optional): description
            options(: obj: `type`, optional): enumeration of options
            default(: obj: `str`, optional): default
            more_info_url(: obj: `str`, optional): URL with more information about the variable
        """
        self.name = name
        self.description = description
        self.options = options
        self.default = default
        self.more_info_url = more_info_url



""" Warnings for logging

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-02-04
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import BioSimulatorsWarning

__all__ = [
    'StandardOutputNotLoggedWarning',
]


class StandardOutputNotLoggedWarning(BioSimulatorsWarning):
    """ Warning that standard output and error could not be logged """
    pass  # pragma: no cover



""" Common environment variables for simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
from ..omex_meta.data_model import OmexMetadataInputFormat, OmexMetadataOutputFormat, OmexMetadataSchema
from ..report.data_model import ReportFormat
from ..viz.data_model import VizFormat
from .data_model import EnvironmentVariable
from kisao import AlgorithmSubstitutionPolicy, ALGORITHM_SUBSTITUTION_POLICY_LEVELS
from unittest import mock

__all__ = [
    'ENVIRONMENT_VARIABLES',
]

with mock.patch.dict('os.environ', {}):
    config = get_config()

ENVIRONMENT_VARIABLES = {
    # formats
    'OMEX_METADATA_INPUT_FORMAT': EnvironmentVariable(
        name='OMEX_METADATA_INPUT_FORMAT',
        description='Which input format to use to validate OMEX metadata files during the validation of COMBINE/OMEX archives.',
        options=sorted(OmexMetadataInputFormat.__members__.keys()),
        default=config.OMEX_METADATA_INPUT_FORMAT.value,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'OMEX_METADATA_OUTPUT_FORMAT': EnvironmentVariable(
        name='OMEX_METADATA_OUTPUT_FORMAT',
        description='Which output format to use to export OMEX metadata files.',
        options=sorted(OmexMetadataOutputFormat.__members__.keys()),
        default=config.OMEX_METADATA_OUTPUT_FORMAT.value,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'OMEX_METADATA_SCHEMA': EnvironmentVariable(
        name='OMEX_METADATA_SCHEMA',
        description='Which schema to use to validate OMEX metadata files during the validation of COMBINE/OMEX archives.',
        options=sorted(OmexMetadataSchema.__members__.keys()),
        default=config.OMEX_METADATA_SCHEMA.value,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    # validation
    'VALIDATE_OMEX_MANIFESTS': EnvironmentVariable(
        name='VALIDATE_OMEX_MANIFESTS',
        description='Whether to validate OMEX manifests during the validation of COMBINE/OMEX archives.',
        options=['0', '1'],
        default='1' if config.VALIDATE_OMEX_MANIFESTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VALIDATE_SEDML': EnvironmentVariable(
        name='VALIDATE_SEDML',
        description='Whether to validate SED-ML files during the validation of COMBINE/OMEX archives.',
        options=['0', '1'],
        default='1' if config.VALIDATE_SEDML else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VALIDATE_SEDML_MODELS': EnvironmentVariable(
        name='VALIDATE_SEDML_MODELS',
        description='Whether to validate models referenced by SED-ML files during the validation of COMBINE/OMEX archives.',
        options=['0', '1'],
        default='1' if config.VALIDATE_SEDML_MODELS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VALIDATE_OMEX_METADATA': EnvironmentVariable(
        name='VALIDATE_OMEX_METADATA',
        description='Whether to validate OMEX metadata (RDF files) during the validation of COMBINE/OMEX archives.',
        options=['0', '1'],
        default='1' if config.VALIDATE_OMEX_METADATA else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VALIDATE_IMAGES': EnvironmentVariable(
        name='VALIDATE_IMAGES',
        description='Whether to validate the images in COMBINE/OMEX archives during their validation.',
        options=['0', '1'],
        default='1' if config.VALIDATE_IMAGES else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VALIDATE_RESULTS': EnvironmentVariable(
        name='VALIDATE_RESULTS',
        description='Whether to validate the results of simulations following their execution.',
        options=['0', '1'],
        default='1' if config.VALIDATE_RESULTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    # algorithm substitution
    'ALGORITHM_SUBSTITUTION_POLICY': EnvironmentVariable(
        name='ALGORITHM_SUBSTITUTION_POLICY',
        description='Policy for substituting alternative algorithms.',
        options=[policy.value for policy, _ in sorted(ALGORITHM_SUBSTITUTION_POLICY_LEVELS.items(),
                                                      key=lambda policy_level: policy_level[1])],
        default=AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES.value,
        more_info_url='https://biosimulators.org/conventions/simulator-interfaces',
    ),

    # reports
    'COLLECT_COMBINE_ARCHIVE_RESULTS': EnvironmentVariable(
        name='COLLECT_COMBINE_ARCHIVE_RESULTS',
        description='Whether to assemble an in memory data structure with all of the simulation results of COMBINE/OMEX archives.',
        options=['0', '1'],
        default='1' if config.COLLECT_SED_DOCUMENT_RESULTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'COLLECT_SED_DOCUMENT_RESULTS': EnvironmentVariable(
        name='COLLECT_SED_DOCUMENT_RESULTS',
        description='Whether to assemble an in memory data structure with all of the simulation results of SED documents.',
        options=['0', '1'],
        default='1' if config.COLLECT_SED_DOCUMENT_RESULTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'REPORT_FORMATS': EnvironmentVariable(
        name='REPORT_FORMATS',
        description='Comma-separated list of formats to save each SED-ML report.',
        options=sorted(ReportFormat.__members__.keys()),
        default=', '.join(config.REPORT_FORMATS),
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'SAVE_PLOT_DATA': EnvironmentVariable(
        name='SAVE_PLOT_DATA',
        description='Whether to save the data for each SED-ML plot similar to a SED-ML report.',
        options=['0', '1'],
        default='1' if config.SAVE_PLOT_DATA else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'H5_REPORTS_PATH': EnvironmentVariable(
        name='H5_REPORTS_PATH',
        description=(
            'Path relative to output directories to export the results of each SED-ML report '
            '(and, optionally, each plot) in HDF5 format.'
        ),
        options=None,
        default=config.H5_REPORTS_PATH,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'REPORTS_PATH': EnvironmentVariable(
        name='REPORTS_PATH',
        description=(
            'Path relative to output directories to export a zip archive with files that represent '
            'the results of each SED-ML report (and, optionally, each plot) in formats other than HDF5.'
        ),
        options=None,
        default=config.REPORTS_PATH,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    # plots
    'VIZ_FORMATS': EnvironmentVariable(
        name='VIZ_FORMATS',
        description='Comma-separated list of formats to save each SED-ML plot.',
        options=sorted(VizFormat.__members__.keys()),
        default=', '.join(config.VIZ_FORMATS),
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'PLOTS_PATH': EnvironmentVariable(
        name='PLOTS_PATH',
        description=(
            'Path relative to output directories to export a zip archive with files for each SED-ML plot.'
        ),
        options=None,
        default=config.PLOTS_PATH,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    # bundling plots and reports
    'BUNDLE_OUTPUTS': EnvironmentVariable(
        name='BUNDLE_OUTPUTS',
        description=(
            'Whether to bundle all of the exported reports and plots generated by '
            'the execution of a COMBINE/OMEX archive into zip files for all reports and all plots.'
        ),
        options=['0', '1'],
        default='1' if config.BUNDLE_OUTPUTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'KEEP_INDIVIDUAL_OUTPUTS': EnvironmentVariable(
        name='KEEP_INDIVIDUAL_OUTPUTS',
        description=(
            'Whether to delete the individual report and plot files generated by '
            'the execution of a COMBINE/OMEX after bundling the files into a zip file.'
        ),
        options=['0', '1'],
        default='1' if config.KEEP_INDIVIDUAL_OUTPUTS else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    # logs
    'LOG_PATH': EnvironmentVariable(
        name='LOG_PATH',
        description=(
            'Path relative to output directories to save YAML-formatted logs of the '
            'execution of COMBINE/OMEX archives and SED-ML files.'
        ),
        options=None,
        default=config.LOG_PATH,
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),

    'VERBOSE': EnvironmentVariable(
        name='VERBOSE',
        description='Whether to print additional information about simulation runs.',
        options=['0', '1'],
        default='1' if config.VERBOSE else '0',
        more_info_url='https://biosimulators.org/conventions/simulator-interfaces',
    ),

    # debugging
    'DEBUG': EnvironmentVariable(
        name='DEBUG',
        description='Whether to raise exceptions rather than capturing them.',
        options=['0', '1'],
        default='1' if config.DEBUG else '0',
        more_info_url='https://docs.biosimulators.org/Biosimulators_utils/source/biosimulators_utils.html',
    ),
}
